import asyncio
import contextlib
from pathlib import Path
from typing import List
from unittest import IsolatedAsyncioTestCase

import faker
import pytest
from pydantic import BaseModel

from panther import Panther
from panther.configs import config
from panther.db import Model
from panther.db.connections import db
from panther.exceptions import DatabaseError, NotFoundAPIError

with contextlib.suppress(ImportError):
    # Only required if user wants to use mongodb
    import bson

f = faker.Faker()


class Book(Model):
    name: str


class BookDetail(BaseModel):
    detail: str
    book: Book
    more_books: list[list[Book]]


class Viewer(BaseModel):
    first_name: str


class Library(Model):
    name: str
    books: list[Book]
    viewer: Viewer


class TestBook(Model):
    name: str
    author: str
    pages_count: int


class Author(Model):
    name: str
    books: list[Book]
    books2: List[Book]
    book: Book
    book2: None | Book = None
    book_detail: dict
    our_book_detail: BookDetail


class InvalidModel(Model):
    new_book: dict[str, Book]


class Publisher(Model):
    name: str
    rating: float
    is_active: bool


class _BaseDatabaseTestCase:
    async def test_insert_one(self):
        book = await Book.insert_one(name='my_test')
        author = await Author.insert_one(
            {'name': 'ali'},
            books=[book],
            books2=[book.id, book.model_dump()],
            book=Book(name='test_book1'),
            book2=None,
            book_detail={'book1': book},
            our_book_detail=BookDetail(detail='ok', book=book, more_books=[[book, book]]),
        )

        assert isinstance(book, Book)
        assert book.id
        assert book.name == 'my_test'

        assert author.name == 'ali'
        assert isinstance(author.books, list)
        assert len(author.books) == 1
        assert author.books[0] == book
        assert author.books[0]

        assert isinstance(author.books2, list)
        assert len(author.books2) == 2
        assert author.books2[0] == book
        assert author.books2[1] == book

        assert isinstance(author.book, Book)
        assert author.book.id
        assert author.book.name == 'test_book1'

        assert author.book2 is None

        assert isinstance(author.book_detail, dict)
        assert list(author.book_detail.keys()) == ['book1']
        assert author.book_detail['book1'] == book.id  # Known Issue

        assert isinstance(author.our_book_detail, BookDetail)
        assert author.our_book_detail.detail == 'ok'
        assert author.our_book_detail.book == book
        assert isinstance(author.our_book_detail.more_books, list)
        assert len(author.our_book_detail.more_books) == 1
        assert isinstance(author.our_book_detail.more_books[0], list)
        assert len(author.our_book_detail.more_books[0]) == 2
        assert author.our_book_detail.more_books[0][0] == book
        assert author.our_book_detail.more_books[0][1] == book

    async def test_insert_one_with_model_obj(self):
        book = await Book.insert_one(name='my_test')
        author = await Author.insert_one(
            Author(
                name='ali',
                books=[book],
                books2=[book, book],
                book=Book(name='test_book1'),
                book2=None,
                book_detail={'book1': book},
                our_book_detail=BookDetail(detail='ok', book=book, more_books=[[book, book]]),
            )
        )

        assert isinstance(book, Book)
        assert book.id
        assert book.name == 'my_test'

        assert author.name == 'ali'
        assert isinstance(author.books, list)
        assert len(author.books) == 1
        assert author.books[0] == book
        assert author.books[0]

        assert isinstance(author.books2, list)
        assert len(author.books2) == 2
        assert author.books2[0] == book
        assert author.books2[1] == book

        assert isinstance(author.book, Book)
        assert author.book.id
        assert author.book.name == 'test_book1'

        assert author.book2 is None

        assert isinstance(author.book_detail, dict)
        assert list(author.book_detail.keys()) == ['book1']
        assert author.book_detail['book1'] == book.id  # Known Issue

        assert isinstance(author.our_book_detail, BookDetail)
        assert author.our_book_detail.detail == 'ok'
        assert author.our_book_detail.book == book
        assert isinstance(author.our_book_detail.more_books, list)
        assert len(author.our_book_detail.more_books) == 1
        assert isinstance(author.our_book_detail.more_books[0], list)
        assert len(author.our_book_detail.more_books[0]) == 2
        assert author.our_book_detail.more_books[0][0] == book
        assert author.our_book_detail.more_books[0][1] == book

    async def test_insert_one_with_model_dump(self):
        book = await Book.insert_one(name='my_test')
        author = await Author.insert_one(
            Author(
                name='ali',
                books=[book],
                books2=[book, book],
                book=book,
                book2=None,
                book_detail={'book1': book},
                our_book_detail=BookDetail(detail='ok', book=book, more_books=[[book, book]]),
            ).model_dump()
        )

        assert isinstance(book, Book)
        assert book.id
        assert book.name == 'my_test'

        assert author.name == 'ali'
        assert isinstance(author.books, list)
        assert len(author.books) == 1
        assert author.books[0] == book
        assert author.books[0]

        assert isinstance(author.books2, list)
        assert len(author.books2) == 2
        assert author.books2[0] == book
        assert author.books2[1] == book

        assert isinstance(author.book, Book)
        assert author.book == book

        assert author.book2 is None

        assert isinstance(author.book_detail, dict)
        assert list(author.book_detail.keys()) == ['book1']
        assert author.book_detail['book1'] == book.model_dump()

        assert isinstance(author.our_book_detail, BookDetail)
        assert author.our_book_detail.detail == 'ok'
        assert author.our_book_detail.book == book
        assert isinstance(author.our_book_detail.more_books, list)
        assert len(author.our_book_detail.more_books) == 1
        assert isinstance(author.our_book_detail.more_books[0], list)
        assert len(author.our_book_detail.more_books[0]) == 2
        assert author.our_book_detail.more_books[0][0] == book
        assert author.our_book_detail.more_books[0][1] == book

    async def test_insert_many(self):
        book = await Book.insert_one(name='my_test')
        authors = await Author.insert_many(
            [
                {
                    'name': 'ali',
                    'books': [book],
                    'books2': [book.id, book],
                    'book': Book(name='test_book1'),
                    'book2': None,
                    'book_detail': {'book1': book.model_dump()},
                    'our_book_detail': BookDetail(detail='ok', book=book, more_books=[[book, book]]),
                },
                {
                    'name': 'ali',
                    'books': [book],
                    'books2': [book.id, book],
                    'book': Book(name='test_book2'),
                    'book2': None,
                    'book_detail': {'book1': book.model_dump()},
                    'our_book_detail': BookDetail(detail='ok', book=book, more_books=[[book, book]]),
                },
            ],
        )

        for author in authors:
            assert isinstance(book, Book)
            assert book.id
            assert book.name == 'my_test'

            assert author.name == 'ali'
            assert isinstance(author.books, list)
            assert len(author.books) == 1
            assert author.books[0] == book
            assert author.books[0]

            assert isinstance(author.books2, list)
            assert len(author.books2) == 2
            assert author.books2[0] == book
            assert author.books2[1] == book

            assert isinstance(author.book, Book)
            assert author.book.id
            assert author.book.name in ['test_book1', 'test_book2']

            assert author.book2 is None

            assert isinstance(author.book_detail, dict)
            assert list(author.book_detail.keys()) == ['book1']
            assert author.book_detail['book1'] == book.model_dump()

            assert isinstance(author.our_book_detail, BookDetail)
            assert author.our_book_detail.detail == 'ok'
            assert author.our_book_detail.book == book
            assert isinstance(author.our_book_detail.more_books, list)
            assert len(author.our_book_detail.more_books) == 1
            assert isinstance(author.our_book_detail.more_books[0], list)
            assert len(author.our_book_detail.more_books[0]) == 2
            assert author.our_book_detail.more_books[0][0] == book
            assert author.our_book_detail.more_books[0][1] == book

    async def test_find_one(self):
        book = await Book.insert_one(name='my_test')
        await Author.insert_one(
            name='ali',
            books=[book],
            books2=[book, book],
            book=Book(name='test_book1'),
            book2=None,
            book_detail={'book1': book},
            our_book_detail=BookDetail(detail='ok', book=book, more_books=[[book, book]]),
        )
        author = await Author.find_one(name='ali')

        assert isinstance(book, Book)
        assert book.id
        assert book.name == 'my_test'

        assert author.name == 'ali'
        assert isinstance(author.books, list)
        assert len(author.books) == 1
        assert author.books[0] == book
        assert author.books[0]

        assert isinstance(author.books2, list)
        assert len(author.books2) == 2
        assert author.books2[0] == book
        assert author.books2[1] == book

        assert isinstance(author.book, Book)
        assert author.book.id
        assert author.book.name == 'test_book1'

        assert author.book2 is None

        assert isinstance(author.book_detail, dict)
        assert list(author.book_detail.keys()) == ['book1']
        assert author.book_detail['book1'] == book.id  # Known Issue

        assert isinstance(author.our_book_detail, BookDetail)
        assert author.our_book_detail.detail == 'ok'
        assert author.our_book_detail.book == book
        assert isinstance(author.our_book_detail.more_books, list)
        assert len(author.our_book_detail.more_books) == 1
        assert isinstance(author.our_book_detail.more_books[0], list)
        assert len(author.our_book_detail.more_books[0]) == 2
        assert author.our_book_detail.more_books[0][0] == book
        assert author.our_book_detail.more_books[0][1] == book

    async def test_find(self):
        book = await Book.insert_one(name='my_test')
        await Author.insert_one(
            name='ali',
            books=[book],
            books2=[book, book],
            book=Book(name='test_book1'),
            book2=None,
            book_detail={'book1': book},
            our_book_detail=BookDetail(detail='ok', book=book, more_books=[[book, book]]),
        )
        _authors = await Author.find(name='ali')
        authors = [i for i in _authors]
        assert len(authors) == 1
        author = authors[0]

        assert isinstance(book, Book)
        assert book.id
        assert book.name == 'my_test'

        assert author.name == 'ali'
        assert isinstance(author.books, list)
        assert len(author.books) == 1
        assert author.books[0] == book
        assert author.books[0]

        assert isinstance(author.books2, list)
        assert len(author.books2) == 2
        assert author.books2[0] == book
        assert author.books2[1] == book

        assert isinstance(author.book, Book)
        assert author.book.id
        assert author.book.name == 'test_book1'

        assert author.book2 is None

        assert isinstance(author.book_detail, dict)
        assert list(author.book_detail.keys()) == ['book1']
        assert author.book_detail['book1'] == book.id  # Known Issue

        assert isinstance(author.our_book_detail, BookDetail)
        assert author.our_book_detail.detail == 'ok'
        assert author.our_book_detail.book == book
        assert isinstance(author.our_book_detail.more_books, list)
        assert len(author.our_book_detail.more_books) == 1
        assert isinstance(author.our_book_detail.more_books[0], list)
        assert len(author.our_book_detail.more_books[0]) == 2
        assert author.our_book_detail.more_books[0][0] == book
        assert author.our_book_detail.more_books[0][1] == book

    async def test_insert_one_invalid_model(self):
        book = await Book.insert_one(name='my_test')
        try:
            await InvalidModel.insert_one(new_book={'book': book})
        except DatabaseError as e:
            assert (
                e.args[0]
                == 'Panther does not support dict[str, tests.test_database_advance.Book] as a field type for unwrapping.'
            )

    # New comprehensive test cases
    async def test_find_one_or_insert_existing(self):
        """Test find_one_or_insert when document exists"""
        book = await Book.insert_one(name='existing_book')

        result, is_inserted = await Book.find_one_or_insert(name='existing_book')

        assert not is_inserted
        assert result == book
        assert result.name == 'existing_book'

    async def test_find_one_or_insert_new(self):
        """Test find_one_or_insert when document doesn't exist"""
        result, is_inserted = await Book.find_one_or_insert(name='new_book')

        assert is_inserted
        assert isinstance(result, Book)
        assert result.name == 'new_book'
        assert result.id

    async def test_find_one_or_raise_found(self):
        """Test find_one_or_raise when document exists"""
        book = await Book.insert_one(name='existing_book')

        result = await Book.find_one_or_raise(name='existing_book')

        assert result == book
        assert result.name == 'existing_book'

    async def test_find_one_or_raise_not_found(self):
        """Test find_one_or_raise when document doesn't exist"""
        try:
            await Book.find_one_or_raise(name='non_existent_book')
            assert False, 'Should have raised NotFoundAPIError'
        except NotFoundAPIError as e:
            assert e.detail == 'Book Does Not Exist'

    async def test_exists_true(self):
        """Test exists when document exists"""
        await Book.insert_one(name='existing_book')

        result = await Book.exists(name='existing_book')

        assert result is True

    async def test_exists_false(self):
        """Test exists when document doesn't exist"""
        result = await Book.exists(name='non_existent_book')

        assert result is False

    async def test_cursor_operations(self):
        """Test cursor operations: skip, limit, sort"""
        # Insert multiple books
        books = []
        for i in range(10):
            book = await Book.insert_one(name=f'book_{i}')
            books.append(book)

        # Test skip and limit
        cursor = await Book.find()
        cursor = cursor.skip(2).limit(3)
        results = [book async for book in cursor]

        assert len(results) == 3
        assert results[0].name == 'book_2'
        assert results[1].name == 'book_3'
        assert results[2].name == 'book_4'

    async def test_cursor_sort(self):
        """Test cursor sorting"""
        # Insert books with different names
        await Book.insert_one(name='zebra')
        await Book.insert_one(name='apple')
        await Book.insert_one(name='banana')

        # Test ascending sort
        cursor = await Book.find()
        cursor = cursor.sort([('name', 1)])
        results = [book async for book in cursor]

        assert len(results) == 3
        assert results[0].name == 'apple'
        assert results[1].name == 'banana'
        assert results[2].name == 'zebra'

        # Test descending sort
        cursor = await Book.find()
        cursor = cursor.sort([('name', -1)])
        results = [book async for book in cursor]

        assert len(results) == 3
        assert results[0].name == 'zebra'
        assert results[1].name == 'banana'
        assert results[2].name == 'apple'

    async def test_update_with_complex_data(self):
        """Test update operations with complex nested data"""
        book = await Book.insert_one(name='original_book')
        author = await Author.insert_one(
            name='original_author',
            books=[book],
            books2=[book],
            book=book,
            book2=None,
            book_detail={'book1': book},
            our_book_detail=BookDetail(detail='original', book=book, more_books=[[book]]),
        )

        # Update with new book
        new_book = await Book.insert_one(name='new_book')

        await Author.update_one(
            {'name': 'original_author'},
            name='updated_author',
            books=[new_book],
            book=new_book,
            our_book_detail=BookDetail(detail='updated', book=new_book, more_books=[[new_book]]),
        )

        updated_author = await Author.find_one(name='updated_author')
        assert updated_author.name == 'updated_author'
        assert len(updated_author.books) == 1
        assert updated_author.books[0] == new_book
        assert updated_author.book == new_book
        assert updated_author.our_book_detail.detail == 'updated'

    async def test_bulk_operations(self):
        """Test bulk insert and update operations"""
        # Bulk insert
        books_data = [{'name': 'bulk_book'} for _ in range(5)]
        books = await Book.insert_many(books_data)

        assert len(books) == 5
        for book in books:
            assert book.name == 'bulk_book'

        # Bulk update
        updated_count = await Book.update_many({'name': 'bulk_book'}, name='updated_bulk_book')

        assert updated_count == 5

        # Verify updates
        updated_books = await Book.find(name='updated_bulk_book')
        count = sum(1 for _ in updated_books)
        assert count == 5

    async def test_empty_collection_operations(self):
        """Test operations on empty collections"""
        # Test find on empty collection
        books = await Book.find()
        count = sum(1 for _ in books)
        assert count == 0

        # Test count on empty collection
        count = await Book.count()
        assert count == 0

        # Test find_one on empty collection
        book = await Book.find_one(name='any_name')
        assert book is None

        # Test first/last on empty collection
        first_book = await Book.first()
        assert first_book is None

        last_book = await Book.last()
        assert last_book is None

    async def test_model_validation_errors(self):
        """Test handling of invalid data"""
        # Test with missing required fields
        try:
            await Book.insert_one()  # Missing name field
            assert False, 'Should have raised validation error'
        except DatabaseError as e:
            assert 'Book(name="Field required")' == str(e)

    async def test_complex_nested_queries(self):
        """Test complex nested queries with multiple conditions"""
        # Create test data
        book1 = await Book.insert_one(name='book1')
        book2 = await Book.insert_one(name='book2')

        author1 = await Author.insert_one(
            name='author1',
            books=[book1],
            books2=[book1],
            book=book1,
            book_detail={'book1': book1},
            our_book_detail=BookDetail(detail='detail1', book=book1, more_books=[[book1]]),
        )

        author2 = await Author.insert_one(
            name='author2',
            books=[book2],
            books2=[book2],
            book=book2,
            book_detail={'book2': book2},
            our_book_detail=BookDetail(detail='detail2', book=book2, more_books=[[book2]]),
        )

        # Complex query
        authors = await Author.find(name='author1')
        results = [author for author in authors]
        assert len(results) == 1
        assert results[0].name == 'author1'

    async def test_save_operations(self):
        """Test save operations for both insert and update"""
        # Test save for new object (insert)
        new_book = Book(name='new_book_save')
        await new_book.save()

        assert new_book.id is not None
        assert new_book.name == 'new_book_save'

        # Verify it was actually saved
        saved_book = await Book.find_one(name='new_book_save')
        assert saved_book is not None
        assert saved_book.id == new_book.id

        # Test save for existing object (update)
        new_book.name = 'updated_book_save'
        await new_book.save()

        # Verify update
        updated_book = await Book.find_one(name='updated_book_save')
        assert updated_book is not None
        assert updated_book.id == new_book.id

    async def test_concurrent_operations(self):
        """Test concurrent operations (basic simulation)"""
        # Insert multiple books concurrently
        import asyncio

        async def insert_book(name):
            return await Book.insert_one(name=name)

        # Create multiple concurrent insertions
        tasks = [insert_book(f'concurrent_book_{i}') for i in range(5)]
        books = await asyncio.gather(*tasks)

        assert len(books) == 5
        for i, book in enumerate(books):
            assert book.name == f'concurrent_book_{i}'
            assert book.id is not None

    async def test_large_dataset_operations(self):
        """Test operations with larger datasets"""
        # Insert larger dataset
        books_data = [{'name': f'large_book_{i}'} for i in range(100)]
        books = await Book.insert_many(books_data)

        assert len(books) == 100

        # Test pagination with large dataset
        cursor = await Book.find()
        cursor = cursor.skip(50).limit(25)
        results = [book async for book in cursor]

        assert len(results) == 25
        assert results[0].name == 'large_book_50'
        assert results[-1].name == 'large_book_74'

        # Test count on large dataset
        count = await Book.count()
        assert count >= 100

    async def test_edge_cases(self):
        """Test various edge cases"""
        # Test with empty string
        book = await Book.insert_one(name='')
        assert book.name == ''

        # Test with very long name
        long_name = 'a' * 1000
        book = await Book.insert_one(name=long_name)
        assert book.name == long_name

        # Test with special characters
        special_name = 'book_with_!@#$%^&*()_+-=[]{}|;:,.<>?'
        book = await Book.insert_one(name=special_name)
        assert book.name == special_name

    async def test_relationship_integrity(self):
        """Test relationship integrity and cascading effects"""
        # Create books and authors with relationships
        book1 = await Book.insert_one(name='book1')
        book2 = await Book.insert_one(name='book2')

        author = await Author.insert_one(
            name='test_author',
            books=[book1, book2],
            books2=[book1, book2],
            book=book1,
            book_detail={'book1': book1, 'book2': book2},
            our_book_detail=BookDetail(detail='test', book=book1, more_books=[[book1, book2]]),
        )

        # Verify relationships are properly established
        assert len(author.books) == 2
        assert len(author.books2) == 2
        assert author.book == book1
        assert len(author.our_book_detail.more_books[0]) == 2

    async def test_error_handling_edge_cases(self):
        """Test error handling for various edge cases"""
        # Test with None values where not allowed
        try:
            await Book.insert_one(name=None)
            assert False, 'Should have raised validation error'
        except Exception:
            pass  # Expected to fail

        # Test with invalid data types
        try:
            await Book.insert_one(name=123)  # Should be string
            assert False, 'Should have raised validation error'
        except Exception:
            pass  # Expected to fail

    async def test_performance_optimizations(self):
        """Test performance-related operations"""
        # Test batch operations
        batch_size = 50
        books_data = [{'name': f'batch_book_{i}'} for i in range(batch_size)]

        start_time = asyncio.get_event_loop().time()
        books = await Book.insert_many(books_data)
        end_time = asyncio.get_event_loop().time()

        assert len(books) == batch_size
        # Basic performance check (should complete in reasonable time)
        assert end_time - start_time < 0.1  # Should complete within 0.1 seconds

    async def test_data_consistency(self):
        """Test data consistency across operations"""
        # Insert book
        book = await Book.insert_one(name='consistency_test')

        # Verify it exists
        found_book = await Book.find_one(name='consistency_test')
        assert found_book == book

        # Update it
        await Book.update_one({'name': 'consistency_test'}, name='updated_consistency_test')

        # Verify update
        updated_book = await Book.find_one(name='updated_consistency_test')
        assert updated_book is not None
        assert updated_book.id == book.id

        # Verify old name doesn't exist
        old_book = await Book.find_one(name='consistency_test')
        assert old_book is None

    async def test_partial_update(self):
        """Test partial update (patch) only updates specified fields"""
        # Insert a book with all fields
        book = await TestBook.insert_one(name='partial_book', author='original_author', pages_count=123)
        # Partial update: only update author
        await TestBook.update_one({'id': book.id}, author='patched_author')
        # Fetch and check
        updated = await TestBook.find_one(id=book.id)
        assert updated.author == 'patched_author'
        assert updated.name == 'partial_book'
        assert updated.pages_count == 123
        # Partial update: only update pages_count
        await TestBook.update_one({'id': book.id}, pages_count=456)
        updated2 = await TestBook.find_one(id=book.id)
        assert updated2.pages_count == 456
        assert updated2.author == 'patched_author'
        assert updated2.name == 'partial_book'

    async def test_first(self):
        """Test getting the first document matching a filter"""
        await TestBook.insert_one(name='first1', author='a', pages_count=1)
        await TestBook.insert_one(name='first2', author='a', pages_count=2)
        first = await TestBook.first(author='a')
        assert first is not None
        assert first.author == 'a'
        assert first.name in ['first1', 'first2']

    async def test_last(self):
        """Test getting the last document matching a filter"""
        await TestBook.insert_one(name='last1', author='b', pages_count=1)
        await TestBook.insert_one(name='last2', author='b', pages_count=2)
        last = await TestBook.last(author='b')
        assert last is not None
        assert last.author == 'b'
        assert last.name in ['last1', 'last2']

    async def test_save_nested_models(self):
        """Test saving a model with nested BaseModel fields"""
        viewer = Viewer(first_name='Ali')
        book = await Book.insert_one(name='nested_book')
        library = await Library.insert_one(name='Central', books=[book], viewer=viewer)
        assert library.id
        assert library.name == 'Central'
        assert isinstance(library.books, list)
        assert library.books[0] == book
        assert isinstance(library.viewer, Viewer)
        assert library.viewer.first_name == 'Ali'


class TestPantherDB(_BaseDatabaseTestCase, IsolatedAsyncioTestCase):
    DB_PATH = 'test.pdb'

    @classmethod
    def setUpClass(cls) -> None:
        global DATABASE
        DATABASE = {
            'engine': {'class': 'panther.db.connections.PantherDBConnection', 'path': cls.DB_PATH},
        }
        Panther(__name__, configs=__name__, urls={})

    def tearDown(self) -> None:
        db.session.collection('Book').drop()
        db.session.collection('Author').drop()
        db.session.collection('Publisher').drop()
        db.session.collection('TestBook').drop()
        db.session.collection('Library').drop()

    @classmethod
    def tearDownClass(cls):
        config.refresh()
        Path(cls.DB_PATH).unlink(missing_ok=True)


@pytest.mark.mongodb
class TestMongoDB(_BaseDatabaseTestCase, IsolatedAsyncioTestCase):
    DB_NAME = 'test.pdb'

    @classmethod
    def setUpClass(cls) -> None:
        global DATABASE
        DATABASE = {
            'engine': {
                'class': 'panther.db.connections.MongoDBConnection',
                'host': f'mongodb://127.0.0.1:27017/{cls.DB_NAME}',
            },
        }

    def setUp(self):
        Panther(__name__, configs=__name__, urls={})

    def tearDown(self) -> None:
        db.session.drop_collection('Book')
        db.session.drop_collection('Author')
        db.session.drop_collection('Publisher')
        db.session.drop_collection('TestBook')
        db.session.drop_collection('Library')

    @classmethod
    def tearDownClass(cls) -> None:
        config.refresh()

    async def test_insert_one_raw_document(self):
        book = await Book.insert_one(name='my_test')
        author = await Author.insert_one(
            name='ali',
            books=[book.model_dump()],
            books2=[book.id, book],
            book=Book(name='test_book1'),
            book2=None,
            book_detail={'book1': book},
            our_book_detail=BookDetail(detail='ok', book=book, more_books=[[book, book]]),
        )

        document = await db.session['Author'].find_one()
        assert isinstance(document, dict)
        assert list(document.keys()) == [
            '_id',
            'name',
            'books',
            'books2',
            'book',
            'book2',
            'book_detail',
            'our_book_detail',
        ]

        assert document['_id'] == author._id

        assert document['name'] == 'ali'

        assert isinstance(document['books'], list)
        assert len(document['books']) == 1
        assert isinstance(document['books'][0], bson.ObjectId)
        assert document['books'][0] == book._id

        assert isinstance(document['books2'], list)
        assert len(document['books2']) == 2
        assert isinstance(document['books2'][0], bson.ObjectId)
        assert document['books2'][0] == book._id
        assert isinstance(document['books2'][1], bson.ObjectId)
        assert document['books2'][1] == book._id

        assert isinstance(document['book'], bson.ObjectId)
        assert document['book'] != book._id  # A new book

        assert document['book2'] is None

        assert isinstance(document['book_detail'], dict)
        assert list(document['book_detail'].keys()) == ['book1']
        assert isinstance(document['book_detail']['book1'], bson.ObjectId)
        assert document['book_detail']['book1'] == book._id

        assert isinstance(document['our_book_detail'], dict)
        assert list(document['our_book_detail'].keys()) == ['detail', 'book', 'more_books']
        assert document['our_book_detail']['detail'] == 'ok'
        assert isinstance(document['our_book_detail']['book'], bson.ObjectId)
        assert document['our_book_detail']['book'] == book._id
        assert isinstance(document['our_book_detail']['more_books'], list)
        assert len(document['our_book_detail']['more_books']) == 1
        assert isinstance(document['our_book_detail']['more_books'][0], list)
        assert len(document['our_book_detail']['more_books'][0]) == 2
        assert isinstance(document['our_book_detail']['more_books'][0][0], bson.ObjectId)
        assert document['our_book_detail']['more_books'][0][0] == book._id
        assert isinstance(document['our_book_detail']['more_books'][0][1], bson.ObjectId)
        assert document['our_book_detail']['more_books'][0][1] == book._id

    async def test_aggregation_pipeline(self):
        """Test MongoDB aggregation pipeline"""
        # Insert test data
        publishers = [
            {'name': 'Publisher A', 'rating': 4.5, 'is_active': True},
            {'name': 'Publisher B', 'rating': 3.8, 'is_active': True},
            {'name': 'Publisher C', 'rating': 4.2, 'is_active': False},
            {'name': 'Publisher D', 'rating': 4.7, 'is_active': True},
        ]

        await Publisher.insert_many(publishers)

        # Test aggregation pipeline
        pipeline = [{'$match': {'is_active': True}}, {'$group': {'_id': None, 'avg_rating': {'$avg': '$rating'}}}]

        results = await Publisher.aggregate(pipeline)
        results_list = list(results)

        assert len(results_list) == 1
        assert 'avg_rating' in results_list[0]
        # Should be average of 4.5, 3.8, and 4.7 (active publishers only)
        expected_avg = (4.5 + 3.8 + 4.7) / 3
        assert abs(results_list[0]['avg_rating'] - expected_avg) < 0.01

    async def test_mongodb_specific_features(self):
        """Test MongoDB-specific features like ObjectId handling"""
        # Test ObjectId conversion
        book = await Book.insert_one(name='mongodb_test')

        # Verify _id property works correctly
        assert hasattr(book, '_id')
        assert isinstance(book._id, bson.ObjectId)
        assert str(book._id) == str(book.id)

        # Test querying by ObjectId
        found_book = await Book.find_one(id=book._id)
        assert found_book == book

        # Test querying by string ID
        found_book_str = await Book.find_one(id=str(book._id))
        assert found_book_str == book

    async def test_mongodb_complex_queries(self):
        """Test complex MongoDB queries with operators"""
        # Insert test data with various ratings
        publishers = [
            {'name': 'High Rated', 'rating': 4.8, 'is_active': True},
            {'name': 'Medium Rated', 'rating': 3.5, 'is_active': True},
            {'name': 'Low Rated', 'rating': 2.1, 'is_active': False},
            {'name': 'Top Rated', 'rating': 4.9, 'is_active': True},
            {'name': 'Top Rated', 'rating': 5.9, 'is_active': False},
        ]

        await Publisher.insert_many(publishers)

        # Test complex query with multiple conditions
        high_rated_active = await Publisher.find({'rating': {'$gte': 4.0}, 'is_active': True})
        results = [pub for pub in high_rated_active]

        assert len(results) == 2  # High Rated, Top Rated, and one more
        for pub in results:
            assert pub.rating >= 4.0
            assert pub.is_active is True

    async def test_mongodb_bulk_operations(self):
        """Test MongoDB-specific bulk operations"""
        # Test bulk write operations
        books_data = [{'name': f'mongodb_bulk_{i}'} for i in range(10)]
        books = await Book.insert_many(books_data)

        # Test bulk update with MongoDB operators
        update_result = await Book.update_many(
            {'name': {'$regex': 'mongodb_bulk_'}}, {'$set': {'name': 'updated_mongodb_bulk'}}
        )

        # Verify all documents were updated
        updated_books = await Book.find(name='updated_mongodb_bulk')
        count = sum(1 for _ in updated_books)
        assert count == 10

    async def test_mongodb_data_types(self):
        """Test MongoDB-specific data types and handling"""
        # Test with various data types that MongoDB handles well
        test_data = {
            'name': 'data_type_test',
            'rating': 4.5,
            'is_active': True,
            'tags': ['fiction', 'adventure'],
            'metadata': {'pages': 300, 'language': 'English'},
        }

        # Create a test model for this
        class TestModel(Model):
            name: str
            rating: float
            is_active: bool
            tags: list[str]
            metadata: dict

        publisher = await TestModel.insert_one(**test_data)

        assert publisher.name == 'data_type_test'
        assert publisher.rating == 4.5
        assert publisher.is_active is True
        assert publisher.tags == ['fiction', 'adventure']
        assert publisher.metadata == {'pages': 300, 'language': 'English'}

    async def test_mongodb_error_handling(self):
        """Test MongoDB-specific error handling"""
        # Test with invalid ObjectId
        try:
            await Book.find_one(id='invalid_object_id')
            # Should handle gracefully
        except Exception as e:
            # Should not crash the application
            assert 'invalid' in str(e).lower() or 'objectid' in str(e).lower()



from pathlib import Path
from typing import ClassVar
from unittest import IsolatedAsyncioTestCase

from pydantic import BaseModel, Field

from panther import Panther
from panther.configs import config
from panther.db import Model
from panther.db.connections import db
from panther.db.models import ID
from panther.response import Response


class OutputModelWithAlias(BaseModel):
    id: ID | None = Field(None, validation_alias='_id', alias='_id')
    name: str


class OutputModelWithPrepare(BaseModel):
    value: int

    async def to_response(self, instance, data):
        # Just return the data with a custom key for test
        return {'custom': data['value']}


class User(Model):
    username: str
    password: str


class UserOutputSerializer(BaseModel):
    id: str
    username: str


class OutputModelWithInstanceCheck(BaseModel):
    value: int
    called_with: ClassVar = []  # class variable to record calls

    async def to_response(self, instance, data):
        # Record the arguments for assertion
        OutputModelWithInstanceCheck.called_with.append((instance, data))
        return {'value': data['value'], 'instance_type': type(instance).__name__}


class OutputModelWithInstanceCheckIterable(BaseModel):
    value: int
    called_with: ClassVar = []

    async def to_response(self, instance, data):
        OutputModelWithInstanceCheckIterable.called_with.append((instance, data))
        return {'value': data['value'], 'instance_type': type(instance).__name__}


class TestResponsesOutputModel(IsolatedAsyncioTestCase):
    DB_PATH = 'test.pdb'

    @classmethod
    def setUpClass(cls) -> None:
        global DATABASE
        DATABASE = {
            'engine': {'class': 'panther.db.connections.PantherDBConnection', 'path': cls.DB_PATH},
        }
        Panther(__name__, configs=__name__, urls={})

    def tearDown(self) -> None:
        db.session.collection('Book').drop()

    @classmethod
    def tearDownClass(cls):
        config.refresh()
        Path(cls.DB_PATH).unlink(missing_ok=True)

    async def test_dict(self):
        resp = Response(data={'_id': 1, 'name': 'foo'})
        await resp.serialize_output(OutputModelWithAlias)
        assert resp.data == {'id': '1', 'name': 'foo'}

    async def test_iterable(self):
        resp = Response(data=[{'_id': 1, 'name': 'foo'}, {'_id': 2, 'name': 'bar'}])
        await resp.serialize_output(OutputModelWithAlias)
        assert resp.data == [{'id': '1', 'name': 'foo'}, {'id': '2', 'name': 'bar'}]

    async def test_to_response(self):
        resp = Response(data={'value': 42})
        await resp.serialize_output(OutputModelWithPrepare)
        assert resp.data == {'custom': 42}

    async def test_iterable_to_response(self):
        resp = Response(data=[{'value': 1}, {'value': 2}])
        await resp.serialize_output(OutputModelWithPrepare)
        assert resp.data == [{'custom': 1}, {'custom': 2}]

    async def test_type_error(self):
        resp = Response(data='not a dict')
        with self.assertRaises(TypeError):
            await resp.serialize_output(OutputModelWithAlias)

    async def test_model(self):
        user = await User.insert_one(username='Ali', password='1234')
        resp = Response(data=user)
        await resp.serialize_output(UserOutputSerializer)
        assert resp.data == {'id': user.id, 'username': 'Ali'}

    async def test_to_response_instance_argument(self):
        OutputModelWithInstanceCheck.called_with.clear()
        resp = Response(data={'value': 123, 'etc': 'ok'})
        await resp.serialize_output(OutputModelWithInstanceCheck)
        # The instance should be the original data dict
        assert resp.data == {'value': 123, 'instance_type': 'dict'}
        assert OutputModelWithInstanceCheck.called_with[0][0] == {'value': 123, 'etc': 'ok'}
        assert OutputModelWithInstanceCheck.called_with[0][1] == {'value': 123}

    async def test_iterable_to_response_instance_argument(self):
        OutputModelWithInstanceCheckIterable.called_with.clear()
        data = [{'value': 1, 'etc': 'ok'}, {'value': 2, 'etc': 'ok'}]
        resp = Response(data=data)
        await resp.serialize_output(OutputModelWithInstanceCheckIterable)
        # The instance should be the original list for each call
        assert resp.data == [
            {'value': 1, 'instance_type': 'dict'},
            {'value': 2, 'instance_type': 'dict'},
        ]
        # Both calls should have the same instance (the dict)
        assert OutputModelWithInstanceCheckIterable.called_with[0][0] == {'value': 1, 'etc': 'ok'}
        assert OutputModelWithInstanceCheckIterable.called_with[1][0] == {'value': 2, 'etc': 'ok'}
        assert OutputModelWithInstanceCheckIterable.called_with[0][1] == {'value': 1}
        assert OutputModelWithInstanceCheckIterable.called_with[1][1] == {'value': 2}

    async def test_model_instance_to_response_argument(self):
        class UserOutputWithInstance(BaseModel):
            id: str
            username: str
            called_with: ClassVar = []

            async def to_response(self, instance, data):
                UserOutputWithInstance.called_with.append((instance, data))
                return {'id': data['id'], 'username': data['username'], 'instance_type': type(instance).__name__}

        UserOutputWithInstance.called_with.clear()
        user = await User.insert_one(username='Ali', password='1234')
        resp = Response(data=user)
        await resp.serialize_output(UserOutputWithInstance)
        assert resp.data['id'] == user.id
        assert resp.data['username'] == 'Ali'
        assert resp.data['instance_type'] == 'User'
        assert UserOutputWithInstance.called_with[0][0] is user
        assert UserOutputWithInstance.called_with[0][1]['id'] == user.id



import asyncio
import logging

from panther._utils import is_function_async
from panther.utils import Singleton

logger = logging.getLogger('panther')


class Event(Singleton):
    _startups = []
    _shutdowns = []

    @classmethod
    def startup(cls, func):
        cls._startups.append(func)

        def wrapper():
            return func()

        return wrapper

    @classmethod
    def shutdown(cls, func):
        cls._shutdowns.append(func)

        def wrapper():
            return func()

        return wrapper

    @classmethod
    async def run_startups(cls):
        for func in cls._startups:
            try:
                if is_function_async(func):
                    await func()
                else:
                    func()
            except Exception as e:
                logger.error(f'{func.__name__}() startup event got error: {e}')

    @classmethod
    def run_shutdowns(cls):
        for func in cls._shutdowns:
            if is_function_async(func):
                try:
                    asyncio.run(func())
                except ModuleNotFoundError:
                    # Error: import of asyncio halted; None in sys.modules
                    #   And as I figured it out, it only happens when we are running with
                    #   gunicorn and Uvicorn workers (-k uvicorn.workers.UvicornWorker)
                    pass
            else:
                func()

    @classmethod
    def clear(cls):
        """Clear all stored events (useful for testing)"""
        cls._startups.clear()
        cls._shutdowns.clear()



import logging
from pathlib import Path

from panther.configs import config

LOGS_DIR = config.BASE_DIR / 'logs'


class FileHandler(logging.FileHandler):
    def __init__(self, filename, mode='a', encoding=None, delay=False, errors=None):
        Path(LOGS_DIR).mkdir(exist_ok=True)
        logging.FileHandler.__init__(self, filename, mode=mode, encoding=encoding, delay=delay, errors=errors)


LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'default': {
            '()': 'uvicorn.logging.DefaultFormatter',
            'fmt': '%(levelprefix)s | %(asctime)s | %(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S',
        },
        'panther_file_formatter': {
            '()': 'uvicorn.logging.DefaultFormatter',
            'fmt': '%(asctime)s | %(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S',
        },
    },
    'handlers': {
        'panther_file': {
            'formatter': 'panther_file_formatter',
            'filename': LOGS_DIR / 'main.log',
            'class': 'panther.logging.FileHandler',
            'delay': True,
        },
        'monitoring_file': {
            'formatter': 'panther_file_formatter',
            'filename': LOGS_DIR / 'monitoring.log',
            'class': 'panther.logging.FileHandler',
            'delay': True,
        },
        'query_file': {
            'formatter': 'panther_file_formatter',
            'filename': LOGS_DIR / 'query.log',
            'class': 'panther.logging.FileHandler',
            'delay': True,
        },
        'default': {
            'formatter': 'default',
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stderr',
        },
    },
    'loggers': {
        'panther': {
            'handlers': ['default', 'panther_file'],
            'level': 'DEBUG',
        },
        'monitoring': {
            'handlers': ['monitoring_file'],
            'level': 'DEBUG',
        },
        'query': {
            'handlers': ['default', 'query_file'],
            'level': 'DEBUG',
        },
    },
}



import random
from pathlib import Path
from unittest import IsolatedAsyncioTestCase

import faker
import pytest
from pantherdb import Cursor as PantherDBCursor
from pydantic import BaseModel

from panther import Panther
from panther.configs import config
from panther.db import Model
from panther.db.connections import db
from panther.db.cursor import Cursor as MongoCursor

f = faker.Faker()


class Book(Model):
    name: str
    author: str
    pages_count: int


class Viewer(BaseModel):
    first_name: str


class Library(Model):
    name: str
    books: list[Book]
    viewer: Viewer


class _BaseDatabaseTestCase:
    # # # Insert
    async def test_insert_one(self):
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        book = await Book.insert_one(name=name, author=author, pages_count=pages_count)

        assert isinstance(book, Book)
        assert book.id
        assert book.name == name
        assert book.pages_count == pages_count

    async def test_insert_many_with_insert_one(self):
        insert_count = await self._insert_many()
        assert insert_count > 1

    async def test_insert_many(self):
        insert_count = random.randint(2, 10)
        initial_books_data = [
            {'name': f.name(), 'author': f.name(), 'pages_count': random.randint(0, 10)} for _ in range(insert_count)
        ]
        books = await Book.insert_many(initial_books_data)
        actual_books = [{'name': book.name, 'author': book.author, 'pages_count': book.pages_count} for book in books]
        expected_books = [
            {'name': book['name'], 'author': book['author'], 'pages_count': book['pages_count']}
            for book in initial_books_data
        ]
        assert len(actual_books) == len(expected_books)
        assert actual_books == expected_books

    # # # FindOne
    async def test_find_one_not_found(self):
        # Insert Many
        await self._insert_many()

        # Find One
        book = await Book.find_one(name='NotFound', author='NotFound', pages_count=0)

        assert book is None

    async def test_find_one_in_many_when_its_last(self):
        # Insert Many
        await self._insert_many()

        # Insert One
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        created_book = await Book.insert_one(name=name, author=author, pages_count=pages_count)

        # Find One
        book = await Book.find_one(name=name, author=author, pages_count=pages_count)

        assert isinstance(book, Book)
        assert book.id
        assert str(book._id) == str(book.id)
        assert book.name == name
        assert book.pages_count == pages_count
        assert created_book == book

    async def test_find_one_in_many_when_its_middle(self):
        # Insert Many
        await self._insert_many()

        # Insert One
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        created_book = await Book.insert_one(name=name, author=author, pages_count=pages_count)

        # Insert Many
        await self._insert_many()

        # Find One
        book = await Book.find_one(name=name, author=author, pages_count=pages_count)

        assert isinstance(book, Book)
        assert book.id
        assert book.name == name
        assert book.pages_count == pages_count
        assert created_book == book

    async def test_first(self):
        # Insert Many
        await self._insert_many()

        # Insert Many With Same Params
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        await self._insert_many_with_specific_params(name=name, author=author, pages_count=pages_count)

        # Find First
        book = await Book.first(name=name, author=author, pages_count=pages_count)

        assert isinstance(book, Book)
        assert book.id
        assert book.name == name
        assert book.pages_count == pages_count

    async def test_first_not_found(self):
        # Insert Many
        await self._insert_many()

        # Find First
        book = await Book.first(name='NotFound', author='NotFound', pages_count=0)

        assert book is None

    async def test_last(self):
        # Insert Many
        await self._insert_many()

        # Insert Many With Same Params
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        await self._insert_many_with_specific_params(name=name, author=author, pages_count=pages_count)
        last_obj = await Book.insert_one(name=name, author=author, pages_count=pages_count)

        # Find One
        book = await Book.last(name=name, author=author, pages_count=pages_count)

        assert isinstance(book, Book)
        assert book.id == last_obj.id
        assert book.name == name
        assert book.pages_count == pages_count

    async def test_last_not_found(self):
        await self._insert_many()

        # Find Last
        book = await Book.last(name='NotFound', author='NotFound', pages_count=0)

        assert book is None

    # # # Find
    async def test_find(self):
        # Insert Many
        await self._insert_many()

        # Insert Many With Specific Name
        name = f.name()
        insert_count = await self._insert_many_with_specific_params(name=name)

        # Find
        books = await Book.find(name=name)
        _len = sum(1 for _ in books)

        if self.__class__.__name__ == 'TestMongoDB':
            assert isinstance(books, MongoCursor)
        else:
            assert isinstance(books, PantherDBCursor)
        assert _len == insert_count
        for book in books:
            assert isinstance(book, Book)
            assert book.name == name

    async def test_find_not_found(self):
        # Insert Many
        await self._insert_many()

        # Find
        books = await Book.find(name='NotFound')
        _len = sum(1 for _ in books)

        if self.__class__.__name__ == 'TestMongoDB':
            assert isinstance(books, MongoCursor)
        else:
            assert isinstance(books, PantherDBCursor)
        assert _len == 0

    async def test_find_without_filter(self):
        # Insert Many
        insert_count = await self._insert_many()

        # Find All
        books = await Book.find()
        _len = sum(1 for _ in books)

        if self.__class__.__name__ == 'TestMongoDB':
            assert isinstance(books, MongoCursor)
        else:
            assert isinstance(books, PantherDBCursor)
        assert _len == insert_count
        for book in books:
            assert isinstance(book, Book)

    async def test_all(self):
        # Insert Many
        insert_count = await self._insert_many()

        # Find All
        books = await Book.all()
        _len = sum(1 for _ in books)

        if self.__class__.__name__ == 'TestMongoDB':
            assert isinstance(books, MongoCursor)
        else:
            assert isinstance(books, PantherDBCursor)

        assert _len == insert_count
        for book in books:
            assert isinstance(book, Book)

    async def test_aggregation(self):
        # Insert Many
        insert_count = await self._insert_many()

        # Find All with aggregate
        books = await Book.aggregate([])
        _len = sum(1 for _ in books)

        assert isinstance(books, list)

        assert _len == insert_count
        for book in books:
            assert isinstance(book, dict)
            assert {*book.keys()} == {'_id', 'name', 'author', 'pages_count'}

    # # # Count
    async def test_count_all(self):
        # Insert Many
        insert_count = await self._insert_many()

        # Count All
        books_count = await Book.count()

        assert isinstance(books_count, int)
        assert books_count == insert_count

    async def test_count_with_filter(self):
        # Insert Many
        await self._insert_many()

        # Insert Many With Specific Name
        name = f.name()
        insert_count = await self._insert_many_with_specific_params(name=name)

        # Count
        books_count = await Book.count(name=name)

        assert isinstance(books_count, int)
        assert books_count == insert_count

    async def test_count_not_found(self):
        # Insert Many
        await self._insert_many()

        # Count
        books_count = await Book.count(name='NotFound')

        assert isinstance(books_count, int)
        assert books_count == 0

    # # # Delete One
    async def test_delete_one(self):
        # Insert Many
        await self._insert_many()

        # Insert With Specific Name
        name = f.name()
        insert_count = await self._insert_many_with_specific_params(name=name)

        # Delete One
        is_deleted = await Book.delete_one(name=name)

        assert isinstance(is_deleted, bool)
        assert is_deleted is True

        # Count Them After Deletion
        assert await Book.count(name=name) == insert_count - 1

    async def test_delete_self(self):
        # Insert Many
        await self._insert_many()

        # Insert With Specific Name
        name = f.name()
        insert_count = await self._insert_many_with_specific_params(name=name)
        # Delete One
        book = await Book.find_one(name=name)
        await book.delete()
        # Count Them After Deletion
        assert await Book.count(name=name) == insert_count - 1

    async def test_delete_one_not_found(self):
        # Insert Many
        insert_count = await self._insert_many()

        # Delete One
        is_deleted = await Book.delete_one(name='_Invalid_Name_')

        assert isinstance(is_deleted, bool)
        assert is_deleted is False

        # Count All
        assert await Book.count() == insert_count

    # # # Delete Many
    async def test_delete_many(self):
        # Insert Many
        pre_insert_count = await self._insert_many()

        # Insert With Specific Name
        name = f.name()
        insert_count = await self._insert_many_with_specific_params(name=name)

        # Delete Many
        deleted_count = await Book.delete_many(name=name)

        assert isinstance(deleted_count, int)
        assert deleted_count == insert_count

        # Count Them After Deletion
        assert await Book.count(name=name) == 0
        assert await Book.count() == pre_insert_count

    async def test_delete_many_not_found(self):
        # Insert Many
        pre_insert_count = await self._insert_many()

        # Delete Many
        name = 'NotFound'
        deleted_count = await Book.delete_many(name=name)

        assert isinstance(deleted_count, int)
        assert deleted_count == 0

        # Count Them After Deletion
        assert await Book.count(name=name) == 0
        assert await Book.count() == pre_insert_count

    # # # Update
    async def test_update_one(self):
        # Insert Many
        insert_count = await self._insert_many()

        # Insert With Specific Name
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        await Book.insert_one(name=name, author=author, pages_count=pages_count)

        # Update One
        new_name = 'New Name'
        is_updated = await Book.update_one({'name': name}, name=new_name)

        assert isinstance(is_updated, bool)
        assert is_updated is True

        book = await Book.find_one(name=new_name)
        assert isinstance(book, Book)
        assert book.author == author
        assert book.pages_count == pages_count

        # Count Them After Update
        assert await Book.count(name=name) == 0
        assert await Book.count() == insert_count + 1

    async def test_update_self(self):
        # Insert Many
        insert_count = await self._insert_many()

        # Insert With Specific Name
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        await Book.insert_one(name=name, author=author, pages_count=pages_count)

        # Update One
        book = await Book.find_one(name=name)
        new_name = 'New Name'
        await book.update(name=new_name)

        assert book.name == new_name
        assert book.author == author

        book = await Book.find_one(name=new_name)
        assert isinstance(book, Book)
        assert book.author == author
        assert book.pages_count == pages_count

        # Count Them After Update
        assert await Book.count(name=name) == 0
        assert await Book.count() == insert_count + 1

    async def test_update_one_not_found(self):
        # Insert Many
        insert_count = await self._insert_many()

        # Update One
        new_name = 'New Name'
        is_updated = await Book.update_one({'name': 'NotFound'}, name=new_name)

        assert isinstance(is_updated, bool)
        assert is_updated is False

        book = await Book.find_one(name=new_name)
        assert book is None

        # Count Them After Update
        assert await Book.count() == insert_count

    # # # Update Many
    async def test_update_many(self):
        # Insert Many
        pre_insert_count = await self._insert_many()

        # Insert With Specific Name
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        insert_count = await self._insert_many_with_specific_params(name=name, author=author, pages_count=pages_count)

        # Update Many
        new_name = 'New Name'
        updated_count = await Book.update_many({'name': name}, name=new_name)

        assert isinstance(updated_count, int)
        assert updated_count == insert_count

        books = await Book.find(name=new_name)
        _len = sum(1 for _ in books)

        if self.__class__.__name__ == 'TestMongoDB':
            assert isinstance(books, MongoCursor)
        else:
            assert isinstance(books, PantherDBCursor)
        assert _len == updated_count == insert_count
        for book in books:
            assert book.author == author
            assert book.pages_count == pages_count

        # Count Them After Update
        assert await Book.count() == pre_insert_count + insert_count

    async def test_update_many_not_found(self):
        # Insert Many
        insert_count = await self._insert_many()

        # Update Many
        new_name = 'New Name'
        updated_count = await Book.update_many({'name': 'NotFound'}, name=new_name)

        assert isinstance(updated_count, int)
        assert updated_count == 0

        book = await Book.find_one(name=new_name)
        assert book is None

        # Count Them After Update
        assert await Book.count() == insert_count

    async def test_save_update(self):
        await self._insert_many()

        # Insert With Specific Name
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)
        book = await Book.insert_one(name=name, author=author, pages_count=pages_count)

        # Save
        assert book.id
        assert book.author == author
        book.author = 'New'
        await book.save()

        # Make sure instance is updated
        assert book.author == 'New'

        # Make sure database is updated
        book = await Book.find_one(name=name)
        assert isinstance(book, Book)
        assert book.author == 'New'
        assert book.pages_count == pages_count

    async def test_save_insert(self):
        name = f.name()
        author = f.name()
        pages_count = random.randint(0, 10)

        book = Book(name=name, author=author, pages_count=pages_count)
        await book.save()

        assert book.id
        assert book.author == author
        assert book.pages_count == pages_count

    async def test_save_nested_models(self):
        library = await Library.insert_one(
            name='Name1', viewer=Viewer(first_name='Ali'), books=[Book(name='Book1', author='Author1', pages_count=1)]
        )
        assert library.id
        assert library.name == 'Name1'
        assert library.viewer.first_name == 'Ali'
        assert library.books[0].name == 'Book1'
        assert library.books[0].author == 'Author1'
        assert library.books[0].pages_count == 1

    @classmethod
    async def _insert_many(cls) -> int:
        insert_count = random.randint(2, 10)

        for _ in range(insert_count):
            await Book.insert_one(name=f.name(), author=f.name(), pages_count=random.randint(0, 10))

        return insert_count

    @classmethod
    async def _insert_many_with_specific_params(
        cls,
        name: str = f.name(),
        author: str = f.name(),
        pages_count: int = random.randint(0, 10),
    ) -> int:
        insert_count = random.randint(2, 10)

        for _ in range(insert_count):
            await Book.insert_one(name=name, author=author, pages_count=pages_count)

        return insert_count


class TestPantherDB(_BaseDatabaseTestCase, IsolatedAsyncioTestCase):
    DB_PATH = 'test.pdb'

    @classmethod
    def setUpClass(cls) -> None:
        global DATABASE
        DATABASE = {'engine': {'class': 'panther.db.connections.PantherDBConnection', 'path': cls.DB_PATH}}
        Panther(__name__, configs=__name__, urls={})

    def tearDown(self) -> None:
        db.session.collection('Book').drop()

    @classmethod
    def tearDownClass(cls):
        config.refresh()
        Path(cls.DB_PATH).unlink(missing_ok=True)

    async def test_aggregation(self):
        pass


@pytest.mark.mongodb
class TestMongoDB(_BaseDatabaseTestCase, IsolatedAsyncioTestCase):
    DB_NAME = 'test.pdb'

    @classmethod
    def setUpClass(cls) -> None:
        global DATABASE
        DATABASE = {
            'engine': {
                'class': 'panther.db.connections.MongoDBConnection',
                'host': f'mongodb://127.0.0.1:27017/{cls.DB_NAME}',
            },
        }

    def setUp(self):
        Panther(__name__, configs=__name__, urls={})

    def tearDown(self) -> None:
        db.session.drop_collection('Book')

    @classmethod
    def tearDownClass(cls):
        config.refresh()



from collections.abc import Callable
from pathlib import Path

from rich import print as rich_print
from rich.console import Console
from rich.progress import ProgressBar
from rich.prompt import Prompt

from panther import version
from panther.cli.template import (
    AUTHENTICATION_PART,
    BASE_DIR_PART,
    DATABASE_MONGODB_PART,
    DATABASE_PANTHERDB_PART,
    IMPORT_COMPLETE_LOAD_ENV_PART,
    IMPORT_LOAD_ENV_PART,
    IMPORT_PATH_PART,
    REDIS_PART,
    SECRET_KEY_PART,
    SINGLE_FILE_TEMPLATE,
    TEMPLATE,
    USER_MODEL_PART,
)
from panther.cli.utils import cli_error


class CreateProject:
    # ERASE_LINE = 100 * ' '
    ERASE_LINE = '\x1b[2K'
    # REMOVE_LAST_LINE = f'\033[1A[{ERASE_LINE}'
    REMOVE_LAST_LINE = f'\x1b[1A{ERASE_LINE}'

    def __init__(self):
        self.console = Console()
        self.input_console = Console(style='bold magenta')
        self.project_name = ''
        self.base_directory = '.'
        self.database = '0'
        self.database_encryption = False
        self.redis = False
        self.authentication = False
        self.monitoring = True
        self.log_queries = True
        self.single_file = False
        self.questions = [
            {
                'field': 'project_name',
                'message': 'Project Name',
                'validation_func': lambda x: x != '',
                'error_message': "'{}' is not valid, Can't be empty.",
            },
            {
                'field': 'base_directory',
                'message': 'Directory (default is ./)',
                'validation_func': self._check_all_directories,
                'error_message': '"{}" directory already exists.',
                'show_validation_error': True,
            },
            {
                'field': 'single_file',
                'message': 'Do you want Single-File project',
                'is_boolean': True,
            },
            {
                'field': 'database',
                'message': '    0: PantherDB (File-Base, No Requirements)\n    1: MongoDB (Required `motor`)\n    2: No Database\nChoose your database (default is 0)',
                'validation_func': lambda x: x in ['0', '1', '2'],
                'error_message': "Invalid choice, '{}' not in ['0', '1', '2']",
            },
            {
                'field': 'redis',
                'message': 'Do you want to use Redis (Required `redis`)',
                'is_boolean': True,
            },
            {
                'field': 'authentication',
                'message': 'Do you want to use JWT Authentication (Required `python-jose`)',
                'is_boolean': True,
            },
        ]
        self.progress_len = len(self.questions)
        self.bar = ProgressBar(total=self.progress_len, width=40)

    def create(self, args: list) -> None:
        # Get Project Name
        if len(args) == 0:
            try:
                self.collect_creation_data()
            except KeyboardInterrupt:
                return self.console.print('\nKeyboardInterrupt', style='bold red')
        else:
            self.project_name = args[0]
            # Get Base Directory
            self.base_directory: str = '.'
            if len(args) > 1:
                self.base_directory = args[1]

            existence = self._check_all_directories(self.base_directory, return_error=True)
            if existence is not True:
                return cli_error(f'"{existence}" directory already exists.')

        self.project_name = self.project_name.lower().replace(' ', '_')
        self.base_directory = self.base_directory.replace(' ', '_')
        template = SINGLE_FILE_TEMPLATE if self.single_file else TEMPLATE

        # Create Base Directory
        if self.base_directory != '.':
            Path(self.base_directory).mkdir()

        for file_name, data in template.items():
            path = f'{self.base_directory}/{file_name}'
            if isinstance(data, str):
                # Create File
                self._create_file(path=path, data=data)
            else:
                # Create Sub Directory
                Path(path).mkdir()

                # Create Files of Sub Directory
                for sub_file_name, sub_data in data.items():
                    inner_path = f'{path}/{sub_file_name}'
                    self._create_file(path=inner_path, data=sub_data)

    def _create_file(self, *, path: str, data: str):
        base_dir_part = BASE_DIR_PART if self.authentication or self.database == '1' else ''
        import_load_env_part = IMPORT_LOAD_ENV_PART if self.authentication else ''
        import_complete_load_env_part = IMPORT_COMPLETE_LOAD_ENV_PART if self.authentication else ''
        import_path_part = IMPORT_PATH_PART if self.authentication or self.database == '1' else ''
        user_model_part = USER_MODEL_PART if self.authentication else ''
        authentication_part = AUTHENTICATION_PART if self.authentication else ''
        secret_key_part = SECRET_KEY_PART if self.authentication else ''
        redis_part = REDIS_PART if self.redis else ''
        if self.database == '0':
            database_part = DATABASE_PANTHERDB_PART
        elif self.database == '1':
            database_part = DATABASE_MONGODB_PART
        else:
            database_part = ''

        data = data.replace('{IMPORT_LOAD_ENV}', import_load_env_part)
        data = data.replace('{IMPORT_COMPLETE_LOAD_ENV}', import_complete_load_env_part)
        data = data.replace('{IMPORT_PATH}', import_path_part)
        data = data.replace('{BASE_DIR}', base_dir_part)
        data = data.replace('{SECRET_KEY}', secret_key_part)
        data = data.replace('{USER_MODEL}', user_model_part)
        data = data.replace('{AUTHENTICATION}', authentication_part)
        data = data.replace('{DATABASE}', database_part)
        data = data.replace('{REDIS}', redis_part)

        data = data.replace('{PROJECT_NAME}', self.project_name)
        data = data.replace('{PANTHER_VERSION}', version())
        with Path(path).open('x', encoding='utf-8') as file:
            file.write(data)

    def collect_creation_data(self):
        self.progress(0)

        for i, question in enumerate(self.questions):
            # Clean Question Data
            field_name = question.pop('field')
            question['default'] = getattr(self, field_name)
            is_boolean = question.pop('is_boolean', False)
            convert_output = str  # Do Nothing
            if is_boolean:
                question['message'] += f' (default is {self._to_str(question["default"])})'
                question['validation_func'] = self._is_boolean
                question['error_message'] = "Invalid choice, '{}' not in ['y', 'n']"
                convert_output = self._to_boolean

            # Check Question Condition
            if 'condition' in question and eval(question.pop('condition')) is False:
                print(flush=True)
            # Ask Question
            else:
                setattr(self, field_name, convert_output(self.ask(**question)))
            self.progress(i + 1)

    def ask(
        self,
        message: str,
        default: str | bool,
        error_message: str,
        validation_func: Callable,
        show_validation_error: bool = False,
    ) -> str:
        value = Prompt.ask(message, console=self.input_console).lower() or default
        while not validation_func(value):
            # Remove the last line, show error message and ask again
            [print(end=self.REMOVE_LAST_LINE, flush=True) for _ in range(message.count('\n') + 1)]
            error = validation_func(value, return_error=True) if show_validation_error else value
            self.console.print(error_message.format(error), style='bold red')
            value = Prompt.ask(message, console=self.input_console).lower() or default
            print(end=self.REMOVE_LAST_LINE, flush=True)
        [print(end=self.REMOVE_LAST_LINE, flush=True) for _ in range(message.count('\n'))]
        return value

    def progress(self, step: int, /, extra_rows: int = 0):
        for i in range(extra_rows + 3 if step else 0):
            print(self.REMOVE_LAST_LINE, flush=True, end='\r')
        self.bar.update(step)

        message = 'Created Successfully' if step == self.progress_len else 'Creating Project'
        rich_print(f'[b]{message:<21}[/b]', end='', flush=True)
        rich_print(self.bar, flush=True)
        print('\n', flush=True)

    @classmethod
    def _to_boolean(cls, _input: str) -> bool:
        return _input in ['y', True]

    @classmethod
    def _is_boolean(cls, _input: str) -> bool:
        return _input in ['y', 'n', True, False]

    @classmethod
    def _to_str(cls, _input: bool) -> str:
        return 'y' if _input else 'n'

    @classmethod
    def _check_all_directories(cls, base_directory: str, return_error: bool = False) -> str | bool:
        """Return False or directory_name means that the directory exist."""
        if base_directory != '.' and Path(base_directory).is_dir():
            return base_directory if return_error else False

        for file_name, data in TEMPLATE.items():
            sub_directory = f'{base_directory}/{file_name}'
            if Path(sub_directory).exists():
                return sub_directory if return_error else False

            if isinstance(data, dict):
                for sub_file_name in data:
                    file_path = f'{sub_directory}/{sub_file_name}'
                    if Path(file_path).exists():
                        return file_path if return_error else False
        return True


create = CreateProject().create



from unittest import IsolatedAsyncioTestCase

from panther import Panther
from panther.app import API
from panther.configs import config
from panther.test import APIClient


@API()
async def simple_api():
    return {'detail': 'ok'}


urls = {'test': simple_api}

ALLOW_ORIGINS = ['*']
ALLOW_METHODS = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE', 'OPTIONS']
ALLOW_HEADERS = ['*']
ALLOW_CREDENTIALS = False
EXPOSE_HEADERS = []
CORS_MAX_AGE = 600
MIDDLEWARES = ['panther.middlewares.cors.CORSMiddleware']


class TestCORSMiddlewares(IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.app = Panther(__name__, configs=__name__, urls=urls)
        cls.client = APIClient(app=cls.app)

    @classmethod
    def tearDownClass(cls) -> None:
        config.refresh()

    async def test_default_cors_headers(self):
        allow_origins = config.ALLOW_ORIGINS
        allow_methods = config.ALLOW_METHODS
        allow_headers = config.ALLOW_HEADERS
        allow_credentials = config.ALLOW_CREDENTIALS
        expose_headers = config.EXPOSE_HEADERS
        res = await self.client.get(path='test', headers={'origin': 'http://example.com'})
        assert res.status_code == 200
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Access-Control-Allow-Methods'] == ', '.join(ALLOW_METHODS)
        assert res.headers['Access-Control-Allow-Headers'] == ', '.join(ALLOW_HEADERS)
        assert res.headers['Access-Control-Max-Age'] == str(CORS_MAX_AGE)
        assert 'Access-Control-Allow-Credentials' not in res.headers
        assert 'Access-Control-Expose-Headers' not in res.headers
        config.ALLOW_ORIGINS = allow_origins
        config.ALLOW_METHODS = allow_methods
        config.ALLOW_HEADERS = allow_headers
        config.ALLOW_CREDENTIALS = allow_credentials
        config.EXPOSE_HEADERS = expose_headers

    async def test_preflight_options(self):
        res = await self.client.options(path='test', headers={'origin': 'http://example.com'})
        assert res.status_code == 204
        assert res.headers['Access-Control-Allow-Origin'] == '*'
        assert res.headers['Access-Control-Allow-Methods'] == ', '.join(ALLOW_METHODS)
        assert res.headers['Access-Control-Allow-Headers'] == ', '.join(ALLOW_HEADERS)
        assert res.headers['Access-Control-Max-Age'] == str(CORS_MAX_AGE)

    async def test_custom_origins(self):
        allow_origins = config.ALLOW_ORIGINS
        config.ALLOW_ORIGINS = ['https://allowed.com']
        res = await self.client.get(path='test', headers={'origin': 'https://allowed.com'})
        assert res.headers['Access-Control-Allow-Origin'] == 'https://allowed.com'
        res = await self.client.get(path='test', headers={'origin': 'https://not-allowed.com'})
        assert res.headers['Access-Control-Allow-Origin'] == 'https://allowed.com'
        config.ALLOW_ORIGINS = allow_origins

    async def test_credentials(self):
        allow_credentials = config.ALLOW_CREDENTIALS
        config.ALLOW_CREDENTIALS = True
        res = await self.client.get(path='test', headers={'origin': 'http://example.com'})
        assert res.headers['Access-Control-Allow-Credentials'] == 'true'
        config.ALLOW_CREDENTIALS = allow_credentials

    async def test_expose_headers(self):
        expose_headers = config.EXPOSE_HEADERS
        config.EXPOSE_HEADERS = ['X-My-Header', 'X-Another-Header']
        res = await self.client.get(path='test', headers={'origin': 'http://example.com'})
        assert res.headers['Access-Control-Expose-Headers'] == 'X-My-Header, X-Another-Header'
        config.EXPOSE_HEADERS = expose_headers



import logging
import time
from abc import abstractmethod
from datetime import datetime, timezone
from typing import Literal

from panther.base_websocket import Websocket
from panther.cli.utils import import_error
from panther.configs import config
from panther.db.connections import redis
from panther.db.models import Model
from panther.exceptions import AuthenticationAPIError
from panther.request import Request
from panther.utils import generate_hash_value_from_string

try:
    from jose import JWTError, jwt
except ModuleNotFoundError as e:
    raise import_error(e, package='python-jose')

logger = logging.getLogger('panther')


class BaseAuthentication:
    @abstractmethod
    async def __call__(self, request: Request | Websocket):
        """Return Instance of User"""
        msg = f'{self.__class__.__name__}.__call__() is not implemented.'
        raise self.exception(msg) from None

    @classmethod
    def exception(cls, message: str | Exception, /) -> type[AuthenticationAPIError]:
        logger.error(f'{cls.__name__} Error: "{message}"')
        return AuthenticationAPIError


class JWTAuthentication(BaseAuthentication):
    """
    Retrieve the Authorization from header
    Example:
        Headers: {'authorization': 'Bearer the_jwt_token'}
    """

    model = None
    keyword = 'Bearer'
    algorithm = 'HS256'
    HTTP_HEADER_ENCODING = 'iso-8859-1'  # RFC5987

    async def __call__(self, request: Request | Websocket) -> Model | None:
        """Authenticate the user based on the JWT token in the Authorization header."""
        auth_header = self.get_authorization_header(request)
        # Set None as `request.user`
        if auth_header is None:
            return None

        token = self.get_token(auth_header=auth_header)
        if redis.is_connected and await self.is_token_revoked(token=token):
            msg = 'User logged out'
            raise self.exception(msg) from None
        payload = await self.decode_jwt(token)
        user = await self.get_user(payload)
        user._auth_token = token
        return user

    @classmethod
    def get_authorization_header(cls, request: Request | Websocket) -> list[str] | None:
        """Retrieve the Authorization header from the request."""
        if auth := request.headers.authorization:
            return auth.split()
        return None

    @classmethod
    def get_token(cls, auth_header):
        if len(auth_header) != 2:
            msg = 'Authorization header must contain 2 parts'
            raise cls.exception(msg) from None

        bearer, token = auth_header

        try:
            token.encode(JWTAuthentication.HTTP_HEADER_ENCODING)
        except UnicodeEncodeError as e:
            raise cls.exception(e) from None

        if bearer.lower() != cls.keyword.lower():
            msg = 'Authorization keyword is not valid'
            raise cls.exception(msg) from None

        return token

    @classmethod
    async def decode_jwt(cls, token: str) -> dict:
        """Decode a JWT token and return the payload."""
        try:
            return jwt.decode(
                token=token,
                key=config.JWT_CONFIG.key,
                algorithms=[config.JWT_CONFIG.algorithm],
            )
        except JWTError as e:
            raise cls.exception(e) from None

    @classmethod
    async def get_user(cls, payload: dict) -> Model:
        """Fetch the user based on the decoded JWT payload from cls.model or config.UserModel"""
        if (user_id := payload.get('user_id')) is None:
            msg = 'Payload does not have `user_id`'
            raise cls.exception(msg)

        user_model = cls.model or config.USER_MODEL
        user = await user_model.find_one(id=user_id)
        if user is None:
            raise cls.exception('User not found')

        return user

    @classmethod
    def encode_jwt(cls, user_id: str, token_type: Literal['access', 'refresh'] = 'access') -> str:
        """Generate a JWT token for a given user ID."""
        issued_at = datetime.now(timezone.utc).timestamp()
        if token_type == 'access':
            expire = issued_at + config.JWT_CONFIG.life_time
        else:
            expire = issued_at + config.JWT_CONFIG.refresh_life_time

        claims = {
            'token_type': token_type,
            'user_id': user_id,
            'iat': issued_at,
            'exp': expire,
        }
        return jwt.encode(
            claims,
            key=config.JWT_CONFIG.key,
            algorithm=config.JWT_CONFIG.algorithm,
        )

    @classmethod
    async def login(cls, user) -> dict:
        """Generate access and refresh tokens for user login."""
        return {
            'access_token': cls.encode_jwt(user_id=user.id),
            'refresh_token': cls.encode_jwt(user_id=user.id, token_type='refresh'),
        }

    @classmethod
    async def logout(cls, user) -> None:
        """Log out a user by revoking their JWT token."""
        payload = await cls.decode_jwt(token=user._auth_token)
        await cls.revoke_token_in_cache(token=user._auth_token, exp=payload['exp'])

    @classmethod
    async def refresh(cls, user):
        if hasattr(user, '_auth_refresh_token'):
            # It happens in CookieJWTAuthentication
            token = user._auth_refresh_token
        else:
            token = user._auth_token

        payload = await cls.decode_jwt(token=token)

        if payload['token_type'] != 'refresh':
            raise cls.exception('Invalid token type; expected `refresh` token.')
        # Revoke after use
        await cls.revoke_token_in_cache(token=token, exp=payload['exp'])

        return await cls.login(user=user)

    @classmethod
    async def revoke_token_in_cache(cls, token: str, exp: int) -> None:
        """Mark the token as revoked in the cache."""
        if redis.is_connected:
            key = generate_hash_value_from_string(token)
            remaining_exp_time = int(exp - time.time())
            await redis.set(key, b'', ex=remaining_exp_time)
        else:
            logger.error('Redis is not connected; token revocation is not effective.')

    @classmethod
    async def is_token_revoked(cls, token: str) -> bool:
        """Check if the token is revoked by looking it up in the cache."""
        key = generate_hash_value_from_string(token)
        return bool(await redis.exists(key))


class QueryParamJWTAuthentication(JWTAuthentication):
    """
    Retrieve the Authorization from query params
    Example:
        https://example.com?authorization=the_jwt_without_bearer
    """

    @classmethod
    def get_authorization_header(cls, request: Request | Websocket) -> str | None:
        if 'authorization' in request.query_params:
            return request.query_params['authorization']
        return None

    @classmethod
    def get_token(cls, auth_header) -> str:
        return auth_header


class CookieJWTAuthentication(JWTAuthentication):
    """
    Retrieve the Authorization from cookies
    Example:
        Cookies: access_token=the_jwt_without_bearer
    """

    async def __call__(self, request: Request | Websocket) -> Model:
        user = await super().__call__(request=request)
        if refresh_token := request.headers.get_cookies().get('refresh_token'):
            # It's used in `cls.refresh()`
            user._auth_refresh_token = refresh_token
        return user

    @classmethod
    def get_authorization_header(cls, request: Request | Websocket) -> str | None:
        if token := request.headers.get_cookies().get('access_token'):
            return token
        return None

    @classmethod
    def get_token(cls, auth_header) -> str:
        return auth_header



import datetime
import time
from unittest import TestCase

import pytest

from panther.background_tasks import BackgroundTask, BackgroundTasks, WeekDay, _background_tasks
from panther.configs import config
from panther.utils import Singleton, timezone_now


@pytest.mark.slow
class TestBackgroundTasks(TestCase):
    def setUp(self):
        self.obj = BackgroundTasks()
        config.BACKGROUND_TASKS = True

    def tearDown(self):
        del Singleton._instances[BackgroundTasks]
        config.BACKGROUND_TASKS = False

    @classmethod
    def tearDownClass(cls):
        config.refresh()

    def test_background_tasks_singleton(self):
        new_obj = BackgroundTasks()
        assert id(self.obj) == id(new_obj)

    def test_initialization(self):
        assert hasattr(self.obj, 'initialize') is True
        assert self.obj._initialized is False

        self.obj.initialize()
        assert self.obj._initialized is True
        assert self.obj.tasks == []

    def test_add_single_task(self):
        def func():
            pass

        task = BackgroundTask(func)
        self.obj.initialize()
        self.obj.add_task(task)
        assert self.obj.tasks == [task]

    def test_add_wrong_task(self):
        def func():
            pass

        self.obj.initialize()
        assert self.obj.tasks == []

        with self.assertLogs() as captured:
            self.obj.add_task(func)

        assert len(captured.records) == 1
        assert (
            captured.records[0].getMessage() == f'`{func.__name__}` '
            f'should be instance of `background_tasks.BackgroundTask`'
        )
        assert self.obj.tasks == []

    def test_add_task_with_false_background_task(self):
        numbers = []

        def func(_numbers):
            _numbers.append(1)

        task = BackgroundTask(func, numbers)
        with self.assertLogs() as captured:
            self.obj.add_task(task)
        assert len(captured.records) == 1
        assert captured.records[0].getMessage() == 'Task will be ignored, `BACKGROUND_TASKS` is not True in `configs`'
        assert self.obj.tasks == []

    def test_add_task_with_args(self):
        numbers = []

        def func(_numbers):
            _numbers.append(1)

        task = BackgroundTask(func, numbers)
        self.obj.initialize()
        self.obj.add_task(task)
        assert self.obj.tasks == [task]
        time.sleep(2)
        assert len(numbers) == 1

    def test_add_task_with_kwargs(self):
        numbers = []

        def func(_numbers):
            _numbers.append(1)

        task = BackgroundTask(func, _numbers=numbers)
        self.obj.initialize()
        self.obj.add_task(task)
        assert self.obj.tasks == [task]
        time.sleep(2)
        assert len(numbers) == 1

    def test_add_task_with_custom_interval(self):
        numbers = []

        def func(_numbers):
            _numbers.append(1)

        task = BackgroundTask(func, numbers).every_seconds().interval(3)
        self.obj.initialize()
        self.obj.add_task(task)
        assert self.obj.tasks == [task]
        time.sleep(4)
        assert len(numbers) == 3

    def test_add_task_with_custom_interval_default_timedelta(self):
        numbers = []

        def func(_numbers):
            _numbers.append(1)

        task = BackgroundTask(func, numbers).interval(2)
        self.obj.initialize()
        self.obj.add_task(task)
        assert self.obj.tasks == [task]
        time.sleep(3)
        # default timedelta is 1 minute, so it can't complete the intervals in 2 seconds
        assert len(numbers) == 1

    def test_add_multiple_tasks(self):
        numbers = []

        def func1(_numbers):
            _numbers.append(1)

        def func2(_numbers):
            _numbers.append(1)

        task2 = BackgroundTask(func1, numbers)
        task1 = BackgroundTask(func2, numbers)
        self.obj.initialize()
        self.obj.add_task(task1)
        self.obj.add_task(task2)
        assert self.obj.tasks == [task1, task2]
        assert len(numbers) == 0
        time.sleep(2)
        assert len(numbers) == 2

    def test_concurrent_tasks(self):
        numbers = []

        def func1(_numbers):
            time.sleep(3)
            _numbers.append(1)

        def func2(_numbers):
            time.sleep(3)
            _numbers.append(1)

        _background_tasks.initialize()
        BackgroundTask(func1, numbers).submit()
        BackgroundTask(func2, numbers).submit()
        assert len(numbers) == 0
        time.sleep(5)  # Initialization takes 2 second
        assert len(numbers) == 2

    def test_add_task_with_custom_every_3_seconds(self):
        numbers = []

        def func(_numbers):
            _numbers.append(1)

        task = BackgroundTask(func, numbers).every_seconds(3).interval(2)
        self.obj.initialize()
        self.obj.add_task(task)
        assert self.obj.tasks == [task]
        time.sleep(3)
        assert len(numbers) == 1
        time.sleep(3)
        assert len(numbers) == 2

    def test_submit_single_task(self):
        numbers = []

        def func(_numbers):
            _numbers.append(1)

        _background_tasks.initialize()
        task = BackgroundTask(func, numbers).submit()
        assert _background_tasks.tasks == [task]
        assert len(numbers) == 0
        time.sleep(2)
        assert len(numbers) == 1

    def test_task_at_3second_later(self):
        numbers = []

        def func(_numbers):
            _numbers.append(1)

        now = timezone_now()

        _background_tasks.initialize()
        BackgroundTask(func, numbers).at((now + datetime.timedelta(seconds=3)).time()).submit()
        time.sleep(2)
        assert len(numbers) == 0
        time.sleep(2)
        assert len(numbers) == 1

    def test_task_on_specific_weekday(self):
        numbers = []

        def func(_numbers):
            _numbers.append(1)

        now = timezone_now()
        week_days = {
            0: WeekDay.MONDAY,
            1: WeekDay.TUESDAY,
            2: WeekDay.WEDNESDAY,
            3: WeekDay.THURSDAY,
            4: WeekDay.FRIDAY,
            5: WeekDay.SATURDAY,
            6: WeekDay.SUNDAY,
        }

        _background_tasks.initialize()
        BackgroundTask(func, numbers).on(week_days[now.weekday()]).submit()
        assert len(numbers) == 0
        time.sleep(1)
        assert len(numbers) == 1



from panther.panel.views import CreateView, DetailView, HomeView, LoginView, TableView

url_routing = {
    '': HomeView,
    '<index>/': TableView,
    '<index>/create/': CreateView,
    'login/': LoginView,
    '<index>/<document_id>/': DetailView,
}



from __future__ import annotations

from sys import version_info

from panther.utils import run_coroutine

try:
    from pymongo.cursor import Cursor as _Cursor
except ImportError:
    # This '_Cursor' is not going to be used,
    #   If user really wants to use it,
    #   we are going to force him to install it in `panther.db.connections.MongoDBConnection.init`
    _Cursor = type('_Cursor', (), {})

if version_info >= (3, 11):
    from typing import Self
else:
    from typing import TypeVar

    Self = TypeVar('Self', bound='BaseMongoDBQuery')


class Cursor(_Cursor):
    models = {}

    def __init__(self, collection, *args, cls=None, **kwargs):
        # cls.__name__ and collection.name are equal.
        if cls:
            self.models[collection.name] = cls
            self.cls = cls
            self.filter = kwargs['filter']
        else:
            self.cls = self.models[collection.name]
        super().__init__(collection, *args, **kwargs)

    def __aiter__(self) -> Self:
        return self

    def __iter__(self) -> Self:
        return self

    async def next(self) -> Self:
        return await self.cls._create_model_instance(document=super().next())

    async def __anext__(self) -> Self:
        try:
            return await self.cls._create_model_instance(document=super().next())
        except StopIteration:
            raise StopAsyncIteration

    def __next__(self) -> Self:
        try:
            return run_coroutine(self.cls._create_model_instance(document=super().next()))
        except StopIteration:
            raise

    def __getitem__(self, index: int | slice) -> Cursor[Self] | Self:
        document = super().__getitem__(index)
        if isinstance(document, dict):
            return run_coroutine(self.cls._create_model_instance(document=document))
        return document



import contextlib
import sys

import uvicorn

from panther.cli.utils import cli_error


def run(args: list[str]) -> None:
    try:
        with contextlib.suppress(KeyboardInterrupt):
            # First arg will be ignored by @Click, so ...
            sys.argv = ['main'] + args
            uvicorn.main()
    except TypeError as e:
        cli_error(e)



from __future__ import annotations

import typing
from sys import version_info

from panther.db.connections import db
from panther.db.cursor import Cursor
from panther.db.queries.base_queries import BaseQuery
from panther.db.utils import prepare_id_for_query

if typing.TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

try:
    from bson.codec_options import CodecOptions
    from pymongo.results import InsertManyResult, InsertOneResult
except ImportError:
    # MongoDB-related libraries are not required by default.
    # If the user intends to use MongoDB, they must install the required dependencies explicitly.
    # This will be enforced in `panther.db.connections.MongoDBConnection.init`.
    CodecOptions = type('CodecOptions', (), {})
    InsertOneResult = type('InsertOneResult', (), {})
    InsertManyResult = type('InsertManyResult', (), {})

if version_info >= (3, 11):
    from typing import Self
else:
    from typing import TypeVar

    Self = TypeVar('Self', bound='BaseMongoDBQuery')


class BaseMongoDBQuery(BaseQuery):
    @classmethod
    def _merge(cls, *args, is_mongo: bool = True) -> dict:
        return super()._merge(*args, is_mongo=is_mongo)

    # TODO: https://jira.mongodb.org/browse/PYTHON-4192
    # @classmethod
    # def collection(cls):
    #     return db.session.get_collection(name=cls.__name__, codec_options=CodecOptions(document_class=cls))

    # # # # # Find # # # # #
    @classmethod
    async def find_one(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        if document := await db.session[cls.__name__].find_one(cls._merge(_filter, kwargs)):
            return await cls._create_model_instance(document=document)
        return None

    @classmethod
    async def find(cls, _filter: dict | None = None, /, **kwargs) -> Cursor:
        return Cursor(cls=cls, collection=db.session[cls.__name__].delegate, filter=cls._merge(_filter, kwargs))

    @classmethod
    async def first(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        cursor = await cls.find(_filter, **kwargs)
        async for result in cursor.sort('_id', 1).limit(-1):
            return result
        return None

    @classmethod
    async def last(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        cursor = await cls.find(_filter, **kwargs)
        async for result in cursor.sort('_id', -1).limit(-1):
            return result
        return None

    @classmethod
    async def aggregate(cls, pipeline: Sequence[dict]) -> Iterable[dict]:
        return await db.session[cls.__name__].aggregate(pipeline).to_list(None)

    # # # # # Count # # # # #
    @classmethod
    async def count(cls, _filter: dict | None = None, /, **kwargs) -> int:
        return await db.session[cls.__name__].count_documents(cls._merge(_filter, kwargs))

    # # # # # Insert # # # # #
    @classmethod
    async def insert_one(cls, document: dict) -> Self:
        insert_one_result: InsertOneResult = await db.session[cls.__name__].insert_one(document)
        return insert_one_result.inserted_id

    @classmethod
    async def insert_many(cls, documents: Iterable[dict]) -> list[Self]:
        final_documents = []
        results = []
        for document in documents:
            prepare_id_for_query(document, is_mongo=True)
            final_document = await cls._process_document(document)
            final_documents.append(final_document)
            results.append(await cls._create_model_instance(document=final_document))
        insert_many_result: InsertManyResult = await db.session[cls.__name__].insert_many(final_documents)
        for obj, inserted_id in zip(results, insert_many_result.inserted_ids):
            obj.id = inserted_id
        return results

    # # # # # Delete # # # # #
    @classmethod
    async def delete_one(cls, _filter: dict | None = None, /, **kwargs) -> bool:
        result = await db.session[cls.__name__].delete_one(cls._merge(_filter, kwargs))
        return bool(result.deleted_count)

    @classmethod
    async def delete_many(cls, _filter: dict | None = None, /, **kwargs) -> int:
        result = await db.session[cls.__name__].delete_many(cls._merge(_filter, kwargs))
        return result.deleted_count

    # # # # # Update # # # # #
    @classmethod
    async def update_one(cls, _filter: dict, _update: dict | None = None, /, **kwargs) -> bool:
        prepare_id_for_query(_filter, is_mongo=True)

        # Step 1: Merge document parameters
        # Combine the _update dict with keyword arguments into a single document
        document = cls._merge(_update, kwargs)

        # Step 2: Process and validate document
        # - Validate data types and structure
        # - Convert Model instances to their IDs for database storage
        # - Handle File objects by saving to disk and storing file paths
        # - Process nested objects and relationships
        final_document = await cls._process_document(document)

        # Step 3: Create model instance (Validating)
        # - Retrieve Model instances from database using IDs
        # - Open File objects from their stored paths
        # - Build the complete model instance with all relationships
        await cls._create_model_instance(document=final_document, is_updating=True)

        # Step 4: Create the query
        update_query = {}
        for field, value in final_document.items():
            if field.startswith('$'):
                update_query[field] = value
            else:
                if not '$set' in update_query:
                    update_query['$set'] = {}
                update_query['$set'][field] = value

        result = await db.session[cls.__name__].update_one(_filter, update_query)
        return bool(result.matched_count)

    @classmethod
    async def update_many(cls, _filter: dict, _update: dict | None = None, /, **kwargs) -> int:
        prepare_id_for_query(_filter, is_mongo=True)

        # Step 1: Merge document parameters
        # Combine the _update dict with keyword arguments into a single document
        document = cls._merge(_update, kwargs)

        # Step 2: Process and validate document
        # - Validate data types and structure
        # - Convert Model instances to their IDs for database storage
        # - Handle File objects by saving to disk and storing file paths
        # - Process nested objects and relationships
        final_document = await cls._process_document(document)

        # Step 3: Create model instance (Validating)
        # - Retrieve Model instances from database using IDs
        # - Open File objects from their stored paths
        # - Build the complete model instance with all relationships
        await cls._create_model_instance(document=final_document, is_updating=True)

        # Step 4: Create the query
        update_query = {}
        for field, value in final_document.items():
            if field.startswith('$'):
                update_query[field] = value
            else:
                if not '$set' in update_query:
                    update_query['$set'] = {}
                update_query['$set'][field] = value

        result = await db.session[cls.__name__].update_many(_filter, update_query)
        return result.modified_count



from unittest import IsolatedAsyncioTestCase, TestCase

from panther import Panther
from panther.app import API
from panther.base_websocket import Websocket
from panther.configs import config
from panther.middlewares.base import HTTPMiddleware, WebsocketMiddleware
from panther.request import Request
from panther.test import APIClient, WebsocketClient
from panther.websocket import GenericWebsocket


class MyMiddleware(HTTPMiddleware):
    async def __call__(self, request: Request):
        request.middlewares = [*getattr(request, 'middlewares', []), 'MyMiddleware']
        response = await self.dispatch(request=request)
        response.data = [*response.data, 'MyMiddleware']
        return response


class BeforeMiddleware1(HTTPMiddleware):
    async def __call__(self, request: Request):
        request.middlewares = [*getattr(request, 'middlewares', []), 'BeforeMiddleware1']
        return await self.dispatch(request=request)


class BeforeMiddleware2(HTTPMiddleware):
    async def __call__(self, request: Request):
        request.middlewares = [*getattr(request, 'middlewares', []), 'BeforeMiddleware2']
        return await self.dispatch(request=request)


class AfterMiddleware1(HTTPMiddleware):
    async def __call__(self, request: Request):
        response = await self.dispatch(request=request)
        response.data = [*response.data, 'AfterMiddleware1']
        return response


class AfterMiddleware2(HTTPMiddleware):
    async def __call__(self, request: Request):
        response = await self.dispatch(request=request)
        response.data = [*response.data, 'AfterMiddleware2']
        return response


class AfterMiddleware3(HTTPMiddleware):
    async def __call__(self, request: Request):
        response = await self.dispatch(request=request)
        response.data = [*response.data, 'AfterMiddleware3']
        return response


class MyWSMiddleware1(WebsocketMiddleware):
    async def __call__(self, connection: Websocket):
        connection.middlewares = [*getattr(connection, 'middlewares', []), 'MyWSMiddleware1']
        return await self.dispatch(connection=connection)


class MyWSMiddleware2(WebsocketMiddleware):
    async def __call__(self, connection: Websocket):
        connection.middlewares = [*getattr(connection, 'middlewares', []), 'MyWSMiddleware2']
        return await self.dispatch(connection=connection)


class PrivateMiddleware1(HTTPMiddleware):
    async def __call__(self, request: Request):
        request.middlewares = [*getattr(request, 'middlewares', []), 'PrivateMiddleware1']
        response = await self.dispatch(request=request)
        response.data = [*response.data, 'PrivateMiddleware1']
        return response


class PrivateMiddleware2(HTTPMiddleware):
    async def __call__(self, request: Request):
        request.middlewares = [*getattr(request, 'middlewares', []), 'PrivateMiddleware2']
        response = await self.dispatch(request=request)
        response.data = [*response.data, 'PrivateMiddleware2']
        return response


@API()
async def handle_middlewares(request: Request):
    states = ['FunctionCall']
    if hasattr(request, 'middlewares'):
        states = request.middlewares + states
    return states


@API(middlewares=[])
async def handle_private_empty_middlewares(request: Request):
    states = ['FunctionCall']
    if hasattr(request, 'middlewares'):
        states = request.middlewares + states
    return states


@API(middlewares=[PrivateMiddleware1, PrivateMiddleware2])
async def handle_private_middlewares(request: Request):
    states = ['FunctionCall']
    if hasattr(request, 'middlewares'):
        states = request.middlewares + states
    return states


class WebsocketHandleMiddlewares(GenericWebsocket):
    async def connect(self):
        await self.accept()
        states = ['WebsocketConnect']
        if hasattr(self, 'middlewares'):
            states = self.middlewares + states
        await self.send(states)
        await self.close()


urls = {
    '': handle_middlewares,
    'private-empty': handle_private_empty_middlewares,
    'private': handle_private_middlewares,
    'websocket': WebsocketHandleMiddlewares,
}


class TestMiddleware(IsolatedAsyncioTestCase):
    def tearDown(self):
        config.refresh()

    async def test_before_base_middleware(self):
        global MIDDLEWARES
        MIDDLEWARES = [BeforeMiddleware1]
        app = Panther(__name__, configs=__name__, urls=urls)
        client = APIClient(app=app)
        response = await client.get('')
        assert response.status_code == 200
        assert response.data == ['BeforeMiddleware1', 'FunctionCall']
        MIDDLEWARES = []

    async def test_after_base_middleware(self):
        global MIDDLEWARES
        MIDDLEWARES = [AfterMiddleware1]
        app = Panther(__name__, configs=__name__, urls=urls)
        client = APIClient(app=app)
        response = await client.get('')
        assert response.status_code == 200
        assert response.data == ['FunctionCall', 'AfterMiddleware1']
        MIDDLEWARES = []

    async def test_a_normal_middleware(self):
        global MIDDLEWARES
        MIDDLEWARES = [MyMiddleware]
        app = Panther(__name__, configs=__name__, urls=urls)
        client = APIClient(app=app)
        response = await client.get('')
        assert response.status_code == 200
        assert response.data == ['MyMiddleware', 'FunctionCall', 'MyMiddleware']

    async def test_websocket_middleware_in_http(self):
        global MIDDLEWARES
        MIDDLEWARES = [MyWSMiddleware1]
        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__, urls={})
            except SystemExit:
                assert True
            else:
                assert False
            finally:
                MIDDLEWARES = []

        assert len(captured.records) == 1
        assert (
            captured.records[0].getMessage()
            == "Invalid 'MIDDLEWARES': <class 'tests.test_middlewares.MyWSMiddleware1'> is not a sub class of `HTTPMiddleware`"
        )

    async def test_private_empty_middleware(self):
        app = Panther(__name__, configs=__name__, urls=urls)
        client = APIClient(app=app)
        response = await client.get('private-empty')
        assert response.status_code == 200
        assert response.data == ['FunctionCall']

    async def test_private_middleware(self):
        app = Panther(__name__, configs=__name__, urls=urls)
        client = APIClient(app=app)
        response = await client.get('private')
        assert response.status_code == 200
        assert response.data == [
            'PrivateMiddleware1',
            'PrivateMiddleware2',
            'FunctionCall',
            'PrivateMiddleware2',
            'PrivateMiddleware1',
        ]

    async def test_middlewares_order(self):
        global MIDDLEWARES
        MIDDLEWARES = [
            BeforeMiddleware2,
            AfterMiddleware3,
            MyMiddleware,
            BeforeMiddleware1,
            AfterMiddleware1,
            AfterMiddleware2,
            MyMiddleware,
        ]
        app = Panther(__name__, configs=__name__, urls=urls)
        client = APIClient(app=app)
        response = await client.get('private')
        assert response.status_code == 200
        assert response.data == [
            'BeforeMiddleware2',
            'MyMiddleware',
            'BeforeMiddleware1',
            'MyMiddleware',
            'PrivateMiddleware1',
            'PrivateMiddleware2',
            'FunctionCall',
            'PrivateMiddleware2',
            'PrivateMiddleware1',
            'MyMiddleware',
            'AfterMiddleware2',
            'AfterMiddleware1',
            'MyMiddleware',
            'AfterMiddleware3',
        ]
        MIDDLEWARES = []


class TestWebsocketMiddleware(TestCase):
    @classmethod
    def tearDownClass(cls):
        config.refresh()

    def setUp(self):
        config.HAS_WS = True

    def tearDown(self):
        config.refresh()

    def test_websocket_middleware(self):
        global WS_MIDDLEWARES
        WS_MIDDLEWARES = [MyWSMiddleware1, MyWSMiddleware2]
        app = Panther(__name__, configs=__name__, urls=urls)
        ws = WebsocketClient(app=app)
        responses = ws.connect('websocket')
        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        assert responses[1]['type'] == 'websocket.send'
        assert responses[1]['text'] == '["MyWSMiddleware1","MyWSMiddleware2","WebsocketConnect"]'

        assert responses[2]['type'] == 'websocket.close'
        assert responses[2]['code'] == 1000
        assert responses[2]['reason'] == ''
        WS_MIDDLEWARES = []

    def test_http_middleware_in_websocket(self):
        global WS_MIDDLEWARES
        WS_MIDDLEWARES = [MyWSMiddleware1, MyMiddleware, MyWSMiddleware2]
        with self.assertLogs(level='ERROR') as captured:
            try:
                Panther(name=__name__, configs=__name__, urls={})
            except SystemExit:
                assert True
            else:
                assert False
            finally:
                WS_MIDDLEWARES = []

        assert len(captured.records) == 1
        assert (
            captured.records[0].getMessage()
            == "Invalid 'WS_MIDDLEWARES': <class 'tests.test_middlewares.MyMiddleware'> is not a sub class of `WebsocketMiddleware`"
        )

    def test_base_middleware(self):
        global WS_MIDDLEWARES
        WS_MIDDLEWARES = [MyWSMiddleware1, MyWSMiddleware2]
        app = Panther(__name__, configs=__name__, urls=urls)
        ws = WebsocketClient(app=app)
        responses = ws.connect('websocket')
        assert responses[0]['type'] == 'websocket.accept'
        assert responses[0]['subprotocol'] is None
        assert responses[0]['headers'] == {}

        assert responses[1]['type'] == 'websocket.send'
        assert responses[1]['text'] == '["MyWSMiddleware1","MyWSMiddleware2","WebsocketConnect"]'

        assert responses[2]['type'] == 'websocket.close'
        assert responses[2]['code'] == 1000
        assert responses[2]['reason'] == ''
        WS_MIDDLEWARES = []



from pathlib import Path
from unittest import IsolatedAsyncioTestCase

import pytest

from panther import Panther
from panther.configs import config
from panther.db import Model
from panther.db.connections import db
from panther.generics import CreateAPI, DeleteAPI, ListAPI, RetrieveAPI, UpdateAPI
from panther.pagination import Pagination
from panther.request import Request
from panther.serializer import ModelSerializer
from panther.test import APIClient


class User(Model):
    name: str


class Person(User):
    age: int


class RetrieveAPITest(RetrieveAPI):
    async def get_instance(self, request: Request, **kwargs) -> Model:
        return await User.find_one(id=kwargs['id'])


class ListAPITest(ListAPI):
    async def get_query(self, request: Request, **kwargs):
        return await User.find()


class FullListAPITest(ListAPI):
    sort_fields = ['name', 'age']
    search_fields = ['name']
    filter_fields = ['id', 'name', 'age']
    pagination = Pagination

    async def get_query(self, request: Request, **kwargs):
        return await Person.find()


class UserSerializer(ModelSerializer):
    class Config:
        model = User
        fields = '*'


class UpdateAPITest(UpdateAPI):
    input_model = UserSerializer

    async def get_instance(self, request: Request, **kwargs) -> Model:
        return await User.find_one(id=kwargs['id'])


class CreateAPITest(CreateAPI):
    input_model = UserSerializer


class DeleteAPITest(DeleteAPI):
    async def get_instance(self, request: Request, **kwargs) -> Model:
        return await User.find_one(id=kwargs['id'])


urls = {
    'retrieve/<id>': RetrieveAPITest,
    'list': ListAPITest,
    'full-list': FullListAPITest,
    'update/<id>': UpdateAPITest,
    'create': CreateAPITest,
    'delete/<id>': DeleteAPITest,
}


class _BaseGenericTestCases:
    async def test_retrieve(self):
        user = await User.insert_one(name='Ali')
        res = await self.client.get(f'retrieve/{user.id}')
        assert res.status_code == 200
        assert res.data == {'id': str(user.id), 'name': user.name}

    async def test_list(self):
        users = await User.insert_many([{'name': 'Ali'}, {'name': 'Hamed'}])
        res = await self.client.get('list')
        assert res.status_code == 200
        assert res.data == [{'id': str(u.id), 'name': u.name} for u in users]

    async def test_list_features(self):
        users = await Person.insert_many(
            [
                {'name': 'Ali', 'age': 0},
                {'name': 'Ali', 'age': 1},
                {'name': 'Saba', 'age': 0},
                {'name': 'Saba', 'age': 1},
            ],
        )
        res = await self.client.get('full-list')
        assert res.status_code == 200
        assert set(res.data.keys()) == {'results', 'count', 'previous', 'next'}

        # Normal
        response = [{'name': r['name'], 'age': r['age']} for r in res.data['results']]
        assert response == [
            {'name': 'Ali', 'age': 0},
            {'name': 'Ali', 'age': 1},
            {'name': 'Saba', 'age': 0},
            {'name': 'Saba', 'age': 1},
        ]

        # Sort 1
        res = await self.client.get('full-list', query_params={'sort': '-name'})
        response = [{'name': r['name'], 'age': r['age']} for r in res.data['results']]
        assert response == [
            {'name': 'Saba', 'age': 0},
            {'name': 'Saba', 'age': 1},
            {'name': 'Ali', 'age': 0},
            {'name': 'Ali', 'age': 1},
        ]

        # Sort 2
        res = await self.client.get('full-list', query_params={'sort': '-name,-age'})
        response = [{'name': r['name'], 'age': r['age']} for r in res.data['results']]
        assert response == [
            {'name': 'Saba', 'age': 1},
            {'name': 'Saba', 'age': 0},
            {'name': 'Ali', 'age': 1},
            {'name': 'Ali', 'age': 0},
        ]

        # Sort 3
        res = await self.client.get('full-list', query_params={'sort': 'name,-age'})
        response = [{'name': r['name'], 'age': r['age']} for r in res.data['results']]
        assert response == [
            {'name': 'Ali', 'age': 1},
            {'name': 'Ali', 'age': 0},
            {'name': 'Saba', 'age': 1},
            {'name': 'Saba', 'age': 0},
        ]

        # Filter 1
        res = await self.client.get('full-list', query_params={'sort': 'name,-age', 'name': 'Ali'})
        response = [{'name': r['name'], 'age': r['age']} for r in res.data['results']]
        assert response == [
            {'name': 'Ali', 'age': 1},
            {'name': 'Ali', 'age': 0},
        ]

        # Filter 2
        res = await self.client.get('full-list', query_params={'sort': 'name,-age', 'id': users[1].id})
        response = [{'id': r['id'], 'name': r['name'], 'age': r['age']} for r in res.data['results']]
        assert response == [{'id': str(users[1].id), 'name': 'Ali', 'age': 1}]

        # Filter 3
        res = await self.client.get('full-list', query_params={'sort': 'name,-age', 'name': 'Alex'})
        response = [{'name': r['name'], 'age': r['age']} for r in res.data['results']]
        assert response == []

        # Search
        res = await self.client.get('full-list', query_params={'sort': 'name,-age', 'search': 'Ali'})
        response = [{'name': r['name'], 'age': r['age']} for r in res.data['results']]
        assert response == [
            {'name': 'Ali', 'age': 1},
            {'name': 'Ali', 'age': 0},
        ]

        # Pagination 1
        res = await self.client.get('full-list', query_params={'sort': 'name,-age'})
        assert res.data['previous'] is None
        assert res.data['next'] is None
        assert res.data['count'] == 4

        # Pagination 2
        res = await self.client.get('full-list', query_params={'limit': 2})
        assert res.data['previous'] is None
        assert res.data['next'] == '?limit=2&skip=2'
        assert res.data['count'] == 4
        response = [{'name': r['name'], 'age': r['age']} for r in res.data['results']]
        assert response == [
            {'name': 'Ali', 'age': 0},
            {'name': 'Ali', 'age': 1},
        ]

        res = await self.client.get('full-list', query_params={'limit': 2, 'skip': 2})
        assert res.data['previous'] == '?limit=2&skip=0'
        assert res.data['next'] is None
        assert res.data['count'] == 4
        response = [{'name': r['name'], 'age': r['age']} for r in res.data['results']]
        assert response == [
            {'name': 'Saba', 'age': 0},
            {'name': 'Saba', 'age': 1},
        ]

    async def test_update(self):
        users = await User.insert_many([{'name': 'Ali'}, {'name': 'Hamed'}])
        res = await self.client.put(f'update/{users[1].id}', payload={'name': 'NewName'})
        assert res.status_code == 200
        assert res.data['name'] == 'NewName'

        new_users = await User.find()
        users[1].name = 'NewName'
        assert {(u.id, u.name) for u in new_users} == {(u.id, u.name) for u in users}

    async def test_create(self):
        res = await self.client.post('create', payload={'name': 'Sara'})
        assert res.status_code == 201
        assert res.data['name'] == 'Sara'

        new_users = list(await User.find())
        assert len(new_users)
        assert new_users[0].name == 'Sara'

    async def test_delete(self):
        users = await User.insert_many([{'name': 'Ali'}, {'name': 'Hamed'}])
        res = await self.client.delete(f'delete/{users[1].id}')
        assert res.status_code == 204
        new_users = list(await User.find())
        assert len(new_users) == 1
        assert new_users[0].model_dump() == users[0].model_dump()


class TestPantherDBGeneric(_BaseGenericTestCases, IsolatedAsyncioTestCase):
    DB_PATH = 'test.pdb'

    @classmethod
    def setUpClass(cls) -> None:
        global DATABASE
        DATABASE = {
            'engine': {'class': 'panther.db.connections.PantherDBConnection', 'path': cls.DB_PATH},
        }
        app = Panther(__name__, configs=__name__, urls=urls)
        cls.client = APIClient(app=app)

    def tearDown(self) -> None:
        db.session.collection('User').drop()
        db.session.collection('Person').drop()

    @classmethod
    def tearDownClass(cls):
        config.refresh()
        Path(cls.DB_PATH).unlink(missing_ok=True)


@pytest.mark.mongodb
class TestMongoDBGeneric(_BaseGenericTestCases, IsolatedAsyncioTestCase):
    DB_NAME = 'test.pdb'

    @classmethod
    def setUpClass(cls) -> None:
        global DATABASE
        DATABASE = {
            'engine': {
                'class': 'panther.db.connections.MongoDBConnection',
                'host': f'mongodb://127.0.0.1:27017/{cls.DB_NAME}',
            },
        }

    def setUp(self):
        app = Panther(__name__, configs=__name__, urls=urls)
        self.client = APIClient(app=app)

    def tearDown(self) -> None:
        db.session.drop_collection('User')
        db.session.drop_collection('Person')

    @classmethod
    def tearDownClass(cls):
        config.refresh()



import asyncio
import base64
import hashlib
import logging
import os
from datetime import datetime, timedelta
from pathlib import Path
from threading import Thread
from typing import ClassVar

import pytz

from panther.configs import config

logger = logging.getLogger('panther')

URANDOM_SIZE = 16


class Singleton(object):
    _instances: ClassVar[dict] = {}

    def __new__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__new__(cls)
        return cls._instances[cls]


def load_env(env_file: str | Path, /) -> dict[str, str]:
    variables = {}

    if env_file is None or not Path(env_file).is_file():
        raise ValueError(f'"{env_file}" is not a file.') from None

    with open(env_file) as file:
        for line in file:
            striped_line = line.strip()
            if not striped_line.startswith('#') and '=' in striped_line:
                key, value = striped_line.split('=', 1)
                key = key.strip()
                value = value.strip().strip('"\'')
                if (boolean_value := value.lower()) in ['true', 'false']:
                    variables[key] = bool(boolean_value == 'true')
                else:
                    variables[key] = value

                # Load them as system environment variable
                os.environ[key] = value
    return variables


def generate_secret_key() -> str:
    return base64.urlsafe_b64encode(os.urandom(32)).decode()


def round_datetime(dt: datetime, delta: timedelta) -> datetime:
    """
    Example:
        >>> round_datetime(datetime(2024, 7, 15, 13, 22, 11, 562159), timedelta(days=2))
        datetime.datetime(2024, 7, 16, 0, 0)

        >>> round_datetime(datetime(2024, 7, 16, 13, 22, 11, 562159), timedelta(days=2))
        datetime.datetime(2024, 7, 16, 0, 0)

        >>> round_datetime(datetime(2024, 7, 17, 13, 22, 11, 562159), timedelta(days=2))
        datetime.datetime(2024, 7, 18, 0, 0)

        >>> round_datetime(datetime(2024, 7, 18, 13, 22, 11, 562159), timedelta(days=2))
        datetime.datetime(2024, 7, 18, 0, 0)

    """
    return datetime.min + round((dt - datetime.min) / delta) * delta


def generate_hash_value_from_string(string_value: str, /) -> str:
    # The point of this method is for maintenance, if we want to change
    # the hash algorithm in the future, it will be easy.
    return hashlib.sha256(string_value.encode('utf-8')).hexdigest()


def scrypt(password: str, salt: bytes, digest: bool = False) -> str | bytes:
    """
    n: CPU/memory cost parameter – Must be a power of 2 (e.g. 1024)
    r: Block size parameter, which fine-tunes sequential memory read size and performance. (8 is commonly used)
    p: Parallelization parameter. (1 .. 232-1 * hLen/MFlen)
    dk_len: Desired key length in bytes (
        Intended output length in octets of the derived key; a positive integer satisfying dkLen ≤ (232− 1) * hLen.)
    h_len: The length in octets of the hash function (32 for SHA256).
    mf_len: The length in octets of the output of the mixing function (SMix below). Defined as r * 128 in RFC7914.
    """
    n = 2**14  # 16384
    r = 8
    p = 10
    dk_len = 64

    derived_key = hashlib.scrypt(password=password.encode(), salt=salt, n=n, r=r, p=p, dklen=dk_len)
    if digest:
        return hashlib.md5(derived_key).hexdigest()
    return derived_key


def timezone_now():
    return datetime.now(tz=pytz.timezone(config.TIMEZONE))


def run_coroutine(coroutine):
    try:
        # Check if there's an event loop already running in this thread
        asyncio.get_running_loop()
    except RuntimeError:
        # No event loop is running in this thread — safe to use asyncio.run
        return asyncio.run(coroutine)

    # Since we cannot block a running event loop with run_until_complete,
    # we execute the coroutine in a separate thread with its own event loop.
    result = []

    def run_in_thread():
        new_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(new_loop)
        try:
            result.append(new_loop.run_until_complete(coroutine))
        finally:
            new_loop.close()

    thread = Thread(target=run_in_thread)
    thread.start()
    thread.join()
    return result[0]



import contextlib
import logging
from abc import abstractmethod

from pantherdb import Cursor as PantherDBCursor

from panther import status
from panther.app import GenericAPI
from panther.configs import config
from panther.db import Model
from panther.db.connections import MongoDBConnection
from panther.db.cursor import Cursor
from panther.db.models import ID
from panther.exceptions import APIError
from panther.pagination import Pagination
from panther.request import Request
from panther.response import Response
from panther.serializer import ModelSerializer

with contextlib.suppress(ImportError):
    # Only required if user wants to use mongodb
    import bson

logger = logging.getLogger('panther')


class RetrieveAPI(GenericAPI):
    @abstractmethod
    async def get_instance(self, request: Request, **kwargs) -> Model:
        """
        Should return an instance of Model, e.g. `await User.find_one()`
        """
        logger.error(f'`get_instance()` method is not implemented in {self.__class__} .')
        raise APIError(status_code=status.HTTP_501_NOT_IMPLEMENTED)

    async def get(self, request: Request, **kwargs):
        instance = await self.get_instance(request=request, **kwargs)
        return Response(data=instance, status_code=status.HTTP_200_OK)


class ListAPI(GenericAPI):
    sort_fields: list[str] = []
    search_fields: list[str] = []
    filter_fields: list[str] = []
    pagination: type[Pagination] | None = None

    @abstractmethod
    async def get_query(self, request: Request, **kwargs):
        """
        Should return a Cursor, e.g. `await User.find()`
        """
        logger.error(f'`get_query()` method is not implemented in {self.__class__} .')
        raise APIError(status_code=status.HTTP_501_NOT_IMPLEMENTED)

    async def get(self, request: Request, **kwargs):
        cursor, pagination = await self.prepare_cursor(request=request, **kwargs)
        return Response(data=cursor, pagination=pagination, status_code=status.HTTP_200_OK)

    async def prepare_cursor(self, request: Request, **kwargs) -> tuple[Cursor | PantherDBCursor, Pagination | None]:
        cursor = await self.get_query(request=request, **kwargs)
        if not isinstance(cursor, (Cursor, PantherDBCursor)):
            logger.error(f'`{self.__class__.__name__}.get_query()` should return a Cursor, e.g. `await Model.find()`')
            raise APIError(status_code=status.HTTP_501_NOT_IMPLEMENTED)

        query = {}
        query |= self.process_filters(query_params=request.query_params, cursor=cursor)
        query |= self.process_search(query_params=request.query_params)

        if query:
            cursor = await cursor.cls.find(cursor.filter | query)

        if sort := self.process_sort(query_params=request.query_params):
            cursor = cursor.sort(sort)

        if pagination := self.process_pagination(query_params=request.query_params, cursor=cursor):
            cursor = pagination.paginate()

        return cursor, pagination

    def process_filters(self, query_params: dict, cursor: Cursor | PantherDBCursor) -> dict:
        _filter = {}
        for field in self.filter_fields:
            if field in query_params:
                _filter[field] = query_params[field]
                if isinstance(config.DATABASE, MongoDBConnection) and cursor.cls.model_fields[field].annotation == ID:
                    _filter[field] = bson.ObjectId(_filter[field])
        return _filter

    def process_search(self, query_params: dict) -> dict:
        search_param = query_params.get('search')
        if not self.search_fields or not search_param:
            return {}
        if isinstance(config.DATABASE, MongoDBConnection):
            if search := [{field: {'$regex': search_param}} for field in self.search_fields]:
                return {'$or': search}
        return {field: search_param for field in self.search_fields}

    def process_sort(self, query_params: dict) -> list:
        sort_param = query_params.get('sort')
        if not self.sort_fields or not sort_param:
            return []
        return [
            (field, -1 if param.startswith('-') else 1)
            for param in sort_param.split(',')
            for field in self.sort_fields
            if field == param.removeprefix('-')
        ]

    def process_pagination(self, query_params: dict, cursor: Cursor | PantherDBCursor) -> Pagination | None:
        if self.pagination:
            return self.pagination(query_params=query_params, cursor=cursor)


class CreateAPI(GenericAPI):
    async def post(self, request: Request, **kwargs):
        instance = await request.validated_data.model.insert_one(request.validated_data.model_dump())
        return Response(data=instance, status_code=status.HTTP_201_CREATED)


class UpdateAPI(GenericAPI):
    @abstractmethod
    async def get_instance(self, request: Request, **kwargs) -> Model:
        """
        Should return an instance of Model, e.g. `await User.find_one()`
        """
        logger.error(f'`get_instance()` method is not implemented in {self.__class__} .')
        raise APIError(status_code=status.HTTP_501_NOT_IMPLEMENTED)

    async def put(self, request: Request, **kwargs):
        instance = await self.get_instance(request=request, **kwargs)
        await instance.update(request.validated_data.model_dump())
        return Response(data=instance, status_code=status.HTTP_200_OK)

    async def patch(self, request: Request, **kwargs):
        instance = await self.get_instance(request=request, **kwargs)
        await instance.update(request.validated_data.model_dump(exclude_none=True))
        return Response(data=instance, status_code=status.HTTP_200_OK)


class DeleteAPI(GenericAPI):
    @abstractmethod
    async def get_instance(self, request: Request, **kwargs) -> Model:
        """
        Should return an instance of Model, e.g. `await User.find_one()`
        """
        logger.error(f'`get_instance()` method is not implemented in {self.__class__} .')
        raise APIError(status_code=status.HTTP_501_NOT_IMPLEMENTED)

    async def pre_delete(self, instance, request: Request, **kwargs):
        """Hook for logic before deletion."""
        pass

    async def post_delete(self, instance, request: Request, **kwargs):
        """Hook for logic after deletion."""
        pass

    async def delete(self, request: Request, **kwargs):
        instance = await self.get_instance(request=request, **kwargs)
        await self.pre_delete(instance, request=request, **kwargs)
        await instance.delete()
        await self.post_delete(instance, request=request, **kwargs)
        return Response(status_code=status.HTTP_204_NO_CONTENT)



from pathlib import Path
from unittest import IsolatedAsyncioTestCase

from pydantic import ConfigDict, Field, field_validator

from panther import Panther
from panther.app import API
from panther.configs import config
from panther.db import Model
from panther.db.connections import db
from panther.request import Request
from panther.serializer import ModelSerializer
from panther.test import APIClient

DB_PATH = 'test_db.pdb'
DATABASE = {
    'engine': {'class': 'panther.db.connections.PantherDBConnection', 'path': DB_PATH},
}


class Book(Model):
    name: str
    author: str = Field('default_author')
    pages_count: int = Field(0)


class Author(Model):
    name: str
    book: Book


class NotRequiredFieldsSerializer(ModelSerializer):
    class Config:
        model = Book
        fields = ['author', 'pages_count']


class RequiredFieldsSerializer(ModelSerializer):
    class Config:
        model = Book
        fields = ['name', 'author', 'pages_count']


class OnlyRequiredFieldsSerializer(ModelSerializer):
    class Config:
        model = Book
        fields = ['name', 'author', 'pages_count']
        required_fields = ['author', 'pages_count']


class WithValidatorsSerializer(ModelSerializer):
    class Config:
        model = Book
        fields = ['name', 'author', 'pages_count']
        required_fields = ['author', 'pages_count']

    @field_validator('name', 'author', 'pages_count')
    def validate_other(cls, field):
        return 'validated'

    @field_validator('pages_count')
    def validate_pages_count(cls, field):
        return 100


class WithClassFieldsSerializer(ModelSerializer):
    age: int = Field(10)

    class Config:
        model = Book
        fields = ['name', 'author', 'pages_count']
        required_fields = ['author', 'pages_count']


class WithForeignKeySerializer(ModelSerializer):
    class Config:
        model = Author
        fields = ['name', 'book']


@API(input_model=NotRequiredFieldsSerializer)
async def not_required(request: Request):
    return request.validated_data


@API(input_model=RequiredFieldsSerializer)
async def required(request: Request):
    return request.validated_data


@API(input_model=OnlyRequiredFieldsSerializer)
async def only_required(request: Request):
    return request.validated_data


@API(input_model=WithValidatorsSerializer)
async def with_validators(request: Request):
    return request.validated_data


@API(input_model=WithClassFieldsSerializer)
async def with_class_fields(request: Request):
    return request.validated_data


@API(input_model=WithForeignKeySerializer)
async def with_foreign_key(request: Request):
    return request.validated_data


urls = {
    'not-required': not_required,
    'required': required,
    'only-required': only_required,
    'with-validators': with_validators,
    'class-fields': with_class_fields,
    'foreign-key': with_foreign_key,
}


class TestModelSerializer(IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        app = Panther(__name__, configs=__name__, urls=urls)
        cls.client = APIClient(app=app)

    def tearDown(self) -> None:
        db.session.collection('Book').drop()
        db.session.collection('Author').drop()

    @classmethod
    def tearDownClass(cls) -> None:
        config.refresh()
        Path(DB_PATH).unlink(missing_ok=True)

    # # # Class Usage

    async def test_not_required_fields_empty_response(self):
        payload = {}
        res = await self.client.post('not-required', payload=payload)
        assert res.status_code == 200
        assert res.data == {'author': 'default_author', 'pages_count': 0}

    async def test_not_required_fields_full_response(self):
        payload = {'author': 'ali', 'pages_count': '12'}
        res = await self.client.post('not-required', payload=payload)
        assert res.status_code == 200
        assert res.data == {'author': 'ali', 'pages_count': 12}

    async def test_required_fields_error(self):
        payload = {}
        res = await self.client.post('required', payload=payload)
        assert res.status_code == 400
        assert res.data == {'name': 'Field required'}

    async def test_required_fields_success(self):
        payload = {'name': 'how to code', 'author': 'ali', 'pages_count': '12'}
        res = await self.client.post('required', payload=payload)
        assert res.status_code == 200
        assert res.data == {'name': 'how to code', 'author': 'ali', 'pages_count': 12}

    async def test_only_required_fields_error(self):
        payload = {}
        res = await self.client.post('only-required', payload=payload)
        assert res.status_code == 400
        assert res.data == {'name': 'Field required', 'author': 'Field required', 'pages_count': 'Field required'}

    async def test_only_required_fields_success(self):
        payload = {'name': 'how to code', 'author': 'ali', 'pages_count': '12'}
        res = await self.client.post('only-required', payload=payload)
        assert res.status_code == 200
        assert res.data == {'name': 'how to code', 'author': 'ali', 'pages_count': 12}

    async def test_with_validators(self):
        payload = {'name': 'how to code', 'author': 'ali', 'pages_count': '12'}
        res = await self.client.post('with-validators', payload=payload)
        assert res.status_code == 200
        assert res.data == {'name': 'validated', 'author': 'validated', 'pages_count': 100}

    async def test_with_class_fields_success(self):
        # Test Default Value
        payload1 = {'name': 'how to code', 'author': 'ali', 'pages_count': '12'}
        res = await self.client.post('class-fields', payload=payload1)
        assert res.status_code == 200
        assert res.data == {'name': 'how to code', 'author': 'ali', 'pages_count': 12, 'age': 10}

        # Test Validation
        payload2 = {'name': 'how to code', 'author': 'ali', 'pages_count': '12', 'age': 30}
        res = await self.client.post('class-fields', payload=payload2)
        assert res.status_code == 200
        assert res.data == {'name': 'how to code', 'author': 'ali', 'pages_count': 12, 'age': 30}

    async def test_foreign_key(self):
        book = await Book.insert_one(name='Book1', author='Ali', pages_count=2)
        await Author.insert_one(name='AnotherAli', book=book)
        payload = {'name': 'Something', 'book': book.id}
        res = await self.client.post('foreign-key', payload=payload)
        assert res.status_code == 200
        assert res.data == {
            'name': 'Something',
            'book': {'id': book.id, 'name': 'Book1', 'author': 'Ali', 'pages_count': 2},
        }

    # # # Class Definition

    async def test_define_class_without_meta(self):
        try:

            class Serializer0(ModelSerializer):
                pass
        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == '`class Config` is required in tests.test_serializer.Serializer0.'
        else:
            assert False

    async def test_define_class_without_model(self):
        try:

            class Serializer1(ModelSerializer):
                class Config:
                    pass
        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == '`Serializer1.Config.model` is required.'
        else:
            assert False

    async def test_define_class_without_fields(self):
        try:

            class Serializer2(ModelSerializer):
                class Config:
                    model = Book
        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == '`Serializer2.Config.fields` is required.'
        else:
            assert False

    async def test_define_class_with_invalid_fields(self):
        try:

            class Serializer3(ModelSerializer):
                class Config:
                    model = Book
                    fields = ['ok', 'no']
        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == '`Serializer3.Config.fields.ok` is not in `Book.model_fields`'
        else:
            assert False

    async def test_define_class_with_invalid_required_fields(self):
        try:

            class Serializer4(ModelSerializer):
                class Config:
                    model = Book
                    fields = ['name', 'author']
                    required_fields = ['pages_count']
        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == '`Serializer4.Config.required_fields.pages_count` is not defined in `Config.fields`.'
        else:
            assert False

    async def test_define_class_with_invalid_model(self):
        try:

            class Serializer5(ModelSerializer):
                class Config:
                    model = ModelSerializer
                    fields = ['name', 'author', 'pages_count']
        except Exception as e:
            assert isinstance(e, AttributeError)
            assert (
                e.args[0] == '`Serializer5.Config.model` is not subclass of `panther.db.Model` or `pydantic.BaseModel`.'
            )
        else:
            assert False

    async def test_define_class_with_invalid_exclude_1(self):
        try:

            class Serializer6(ModelSerializer):
                class Config:
                    model = Book
                    fields = ['name', 'author', 'pages_count']
                    exclude = ['not_found']

        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == '`Serializer6.Config.exclude.not_found` is not valid.'
        else:
            assert False

    async def test_define_class_with_invalid_exclude_2(self):
        try:

            class Serializer7(ModelSerializer):
                class Config:
                    model = Book
                    fields = ['name', 'pages_count']
                    exclude = ['author']

        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == '`Serializer7.Config.exclude.author` is not defined in `Config.fields`.'
        else:
            assert False

    async def test_with_star_fields_with_exclude3(self):
        try:

            class Serializer8(ModelSerializer):
                class Config:
                    model = Book
                    fields = ['*']
                    exclude = ['author']

        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == "`Serializer8.Config.fields.*` is not valid. Did you mean `fields = '*'`"
        else:
            assert False

    # # # Serializer Usage
    async def test_with_simple_model_config(self):
        class Serializer(ModelSerializer):
            model_config = ConfigDict(str_to_upper=True)

            class Config:
                model = Book
                fields = ['name', 'author', 'pages_count']

        serialized = Serializer(name='book', author='AliRn', pages_count='12')
        assert serialized.name == 'BOOK'
        assert serialized.author == 'ALIRN'
        assert serialized.pages_count == 12

    async def test_with_inner_model_config(self):
        class Serializer(ModelSerializer):
            class Config:
                str_to_upper = True
                model = Book
                fields = ['name', 'author', 'pages_count']

        serialized = Serializer(name='book', author='AliRn', pages_count='12')
        assert serialized.name == 'BOOK'
        assert serialized.author == 'ALIRN'
        assert serialized.pages_count == 12

    async def test_with_dual_model_config(self):
        class Serializer(ModelSerializer):
            model_config = ConfigDict(str_to_upper=False)

            class Config:
                str_to_upper = True
                model = Book
                fields = ['name', 'author', 'pages_count']

        serialized = Serializer(name='book', author='AliRn', pages_count='12')
        assert serialized.name == 'book'
        assert serialized.author == 'AliRn'
        assert serialized.pages_count == 12

    async def test_serializer_doc(self):
        class Serializer1(ModelSerializer):
            """Hello I'm Doc"""

            class Config:
                model = Book
                fields = ['name', 'author', 'pages_count']

        serialized = Serializer1(name='book', author='AliRn', pages_count='12')
        assert serialized.__doc__ == "Hello I'm Doc"

        class Serializer2(ModelSerializer):
            class Config:
                model = Book
                fields = ['name', 'author', 'pages_count']

        serialized = Serializer2(name='book', author='AliRn', pages_count='12')
        assert serialized.__doc__ is None

    async def test_with_exclude(self):
        class Serializer(ModelSerializer):
            class Config:
                model = Book
                fields = ['name', 'author', 'pages_count']
                exclude = ['author']

        serialized = Serializer(name='book', author='AliRn', pages_count='12')
        assert set(serialized.model_dump().keys()) == {'name', 'pages_count'}
        assert serialized.name == 'book'
        assert serialized.pages_count == 12

    async def test_with_star_fields(self):
        class Serializer(ModelSerializer):
            class Config:
                model = Book
                fields = '*'

        serialized = Serializer(name='book', author='AliRn', pages_count='12')
        assert set(serialized.model_dump().keys()) == {'id', 'name', 'author', 'pages_count'}
        assert serialized.name == 'book'
        assert serialized.author == 'AliRn'
        assert serialized.pages_count == 12

    async def test_with_star_fields_with_exclude(self):
        class Serializer(ModelSerializer):
            class Config:
                model = Book
                fields = '*'
                exclude = ['author']

        serialized = Serializer(name='book', author='AliRn', pages_count='12')
        assert set(serialized.model_dump().keys()) == {'id', 'name', 'pages_count'}
        assert serialized.name == 'book'
        assert serialized.pages_count == 12



from panther.exceptions import RedirectAPIError
from panther.permissions import BasePermission
from panther.request import Request


class IsAuthenticated(BasePermission):
    async def __call__(self, request: Request):
        if request.user is None:
            raise RedirectAPIError(url=f'login?redirect_to={request.path}')



from __future__ import annotations

import operator
import types
import typing
from abc import abstractmethod
from collections.abc import Iterator
from datetime import datetime
from functools import reduce
from sys import version_info
from typing import Any, Union, get_args, get_origin

from pantherdb import Cursor
from pydantic import BaseModel, ValidationError
from pydantic_core._pydantic_core import ValidationError

from panther._utils import detect_mime_type
from panther.db.cursor import Cursor
from panther.db.utils import prepare_id_for_query
from panther.exceptions import DatabaseError
from panther.file_handler import File

if version_info >= (3, 11):
    from typing import Self
else:
    from typing import TypeVar

    Self = TypeVar('Self', bound='BaseQuery')


class BaseQuery:
    @classmethod
    def _merge(cls, *args, is_mongo: bool = False) -> dict:
        prepare_id_for_query(*args, is_mongo=is_mongo)
        return reduce(operator.ior, filter(None, args), {})

    @classmethod
    def _clean_error_message(cls, validation_error: ValidationError, is_updating: bool = False) -> str:
        error = ', '.join(
            '{field}="{error}"'.format(field='.'.join(str(loc) for loc in e['loc']), error=e['msg'])
            for e in validation_error.errors()
            if not is_updating or e['type'] != 'missing'
        )
        return f'{cls.__name__}({error})' if error else ''

    @classmethod
    def _validate_data(cls, *, data: dict, is_updating: bool = False):
        """Validate document before inserting to collection"""
        try:
            cls(**data)
        except ValidationError as validation_error:
            if error := cls._clean_error_message(validation_error=validation_error, is_updating=is_updating):
                raise DatabaseError(error)

    @classmethod
    def _get_annotation_type(cls, annotation: Any) -> type | None:
        """
        Extracts the underlying, non-optional type from a type annotation.
        Handles basic types, Pydantic BaseModels, lists, and unions (optionals).
        Returns None if no single underlying type can be determined (e.g., for list[NoneType]).
        Raises DatabaseError for unsupported annotations.
        """
        origin = get_origin(annotation)

        # Handle list[T] and Union[T, None] (T | None or typing.Union[T, None])
        if origin is list or origin is types.UnionType or origin is Union:
            # Extracts the first non-None type from a tuple of type arguments.
            for arg in get_args(annotation):
                if arg is not type(None):
                    return arg
            return None

        # Handle basic types (str, int, bool, dict, datetime) and Pydantic.BaseModel and File subclasses
        try:
            if isinstance(annotation, type) and (
                annotation in (str, int, bool, dict, float, datetime) or issubclass(annotation, (BaseModel, File))
            ):
                return annotation
        except TypeError:
            # issubclass(dict[str, tests.test_database_advance.Book], BaseModel) in Python3.10 gives us TypeError and
            # works just fine in other versions, so we have to raise DatabaseError in this situation anyway.
            pass
        raise DatabaseError(f'Panther does not support {annotation} as a field type for unwrapping.')

    @classmethod
    async def _create_list(cls, field_type: type, value: Any) -> Any:
        from panther.db import Model

        # `field_type` is the expected type of items in the list (e.g., int, Model, list[str])
        # `value` is a single item from the input list that needs processing.

        # Handles list[list[int]], list[dict[str,int]] etc.
        if isinstance(field_type, (types.GenericAlias, typing._GenericAlias)):
            element_type = cls._get_annotation_type(field_type)  # Unwrap further (e.g. list[str] -> str)
            if element_type is None:
                raise DatabaseError(f'Cannot determine element type for generic list item: {field_type}')
            if not isinstance(value, list):  # Or check if iterable, matching the structure
                raise DatabaseError(f'Expected a list for nested generic type {field_type}, got {type(value)}')
            return [await cls._create_list(field_type=element_type, value=item) for item in value]

        # Make sure Model condition is before BaseModel.
        if isinstance(field_type, type) and issubclass(field_type, Model):
            # `value` is assumed to be an ID for the Model instance.
            return await field_type.first(id=value)

        if isinstance(field_type, type) and issubclass(field_type, BaseModel):
            if not isinstance(value, dict):
                raise DatabaseError(f'Expected a dictionary for BaseModel {field_type.__name__}, got {type(value)}')

            return {
                field_name: await cls._create_field(model=field_type, field_name=field_name, value=value[field_name])
                for field_name in value
            }

        # `value` is a primitive type (str, int, etc.)
        return value

    @classmethod
    async def _create_field(cls, model: type, field_name: str, value: Any) -> Any:
        from panther.db import Model

        # Handle primary key field directly
        if field_name == 'id':
            return value

        if field_name not in model.model_fields:
            # Field from input data is not defined in the model.
            # Pydantic's `extra` config on the model will handle this upon instantiation.
            return value

        field_annotation = model.model_fields[field_name].annotation
        unwrapped_type = cls._get_annotation_type(field_annotation)

        if unwrapped_type is None:
            raise DatabaseError(
                f"Could not determine a valid underlying type for field '{field_name}' "
                f'with annotation {field_annotation} in model {model.__name__}.',
            )

        if get_origin(field_annotation) is list:
            # Or check for general iterables if applicable
            if not isinstance(value, list):
                raise DatabaseError(
                    f"Field '{field_name}' expects a list, got {type(value)} for model {model.__name__}",
                )
            return [await cls._create_list(field_type=unwrapped_type, value=item) for item in value]

        # Condition of `File` should be on top of `BaseModel`
        if isinstance(unwrapped_type, type) and issubclass(unwrapped_type, File):
            # `value` is assumed to be the path of the File.
            content_type = detect_mime_type(file_path=value)
            with open(value, 'rb') as f:
                return File(file_name=value, content_type=content_type, file=f.read()).model_dump()

        # Condition of `Model` should be on top of `BaseModel`
        if isinstance(unwrapped_type, type) and issubclass(unwrapped_type, Model):
            # `value` is assumed to be an ID for the Model instance.
            if obj := await unwrapped_type.first(id=value):
                return obj.model_dump()
            return None

        if isinstance(unwrapped_type, type) and issubclass(unwrapped_type, BaseModel):
            if not isinstance(value, dict):
                raise DatabaseError(
                    f"Field '{field_name}' expects a dictionary for BaseModel {unwrapped_type.__name__}, "
                    f'got {type(value)} in model {model.__name__}',
                )
            return {
                nested_field_name: await cls._create_field(
                    model=unwrapped_type,
                    field_name=nested_field_name,
                    value=value[nested_field_name],
                )
                for nested_field_name in unwrapped_type.model_fields
                if nested_field_name in value
            }

        return value

    @classmethod
    async def _create_model_instance(cls, document: dict, is_updating: bool = False) -> Self:
        """Prepares document and creates an instance of the model."""
        if '_id' in document:
            document['id'] = document.pop('_id')

        processed_document = {
            field_name: await cls._create_field(model=cls, field_name=field_name, value=field_value)
            for field_name, field_value in document.items()
        }
        try:
            return cls(**processed_document)
        except ValidationError as validation_error:
            if error := cls._clean_error_message(validation_error=validation_error, is_updating=is_updating):
                raise DatabaseError(error) from validation_error

    @classmethod
    async def _clean_value(cls, value: Any) -> dict[str, Any] | list[Any]:
        from panther.db import Model

        match value:
            case None:
                return None
            case Model() as model:
                if model.id in [None, '']:
                    await model.save()
                # We save full object because user didn't specify the type.
                return model._id
            case File() as file:
                # Write file to disk
                return file.save()
            case BaseModel() as model:
                return {
                    field_name: await cls._clean_value(value=getattr(model, field_name))
                    for field_name in model.__class__.model_fields
                }
            case dict() as d:
                return {k: await cls._clean_value(value=v) for k, v in d.items()}
            case list() as l:
                return [await cls._clean_value(value=item) for item in l]
        return value

    @classmethod
    async def _extract_type(cls, field_name: str) -> Any:
        if field_name not in cls.model_fields:
            return None
        field_annotation = cls.model_fields[field_name].annotation
        unwrapped_type = cls._get_annotation_type(field_annotation)

        if (
            get_origin(field_annotation) is list
            and isinstance(unwrapped_type, type)
            and issubclass(unwrapped_type, BaseModel)
        ):
            return list[unwrapped_type]

        if isinstance(unwrapped_type, type) and issubclass(unwrapped_type, BaseModel):
            return unwrapped_type

    @classmethod
    async def _process_document(cls, document):
        # 1. Check type expected type and cast to that
        # 2. Check type of field_value and do the stuff (save() or return ._id)
        processed_document = {}
        for field_name, field_value in document.items():
            if field_name in ['id', '_id']:
                continue
            field_type = await cls._extract_type(field_name)
            if field_type:
                if get_origin(field_type) is list:
                    cls_type = cls._get_annotation_type(field_type)
                    value = [cls_type(**v) if isinstance(v, dict) else v for v in field_value]
                else:
                    if isinstance(field_value, dict):
                        value = field_type(**field_value)
                    else:
                        value = field_value
            else:
                value = field_value
            processed_document[field_name] = await cls._clean_value(value)
        return processed_document

    @classmethod
    @abstractmethod
    async def find_one(cls, *args, **kwargs) -> Self | None:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def find(cls, *args, **kwargs) -> list[Self] | Cursor:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def first(cls, *args, **kwargs) -> Self | None:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def last(cls, *args, **kwargs):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def aggregate(cls, *args, **kwargs) -> Iterator[dict]:
        raise NotImplementedError

    # # # # # Count # # # # #
    @classmethod
    @abstractmethod
    async def count(cls, *args, **kwargs) -> int:
        raise NotImplementedError

    # # # # # Insert # # # # #
    @classmethod
    @abstractmethod
    async def insert_one(cls, *args, **kwargs) -> Self:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def insert_many(cls, *args, **kwargs) -> list[Self]:
        raise NotImplementedError

    # # # # # Delete # # # # #
    @abstractmethod
    async def delete(self) -> None:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def delete_one(cls, *args, **kwargs) -> bool:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def delete_many(cls, *args, **kwargs) -> int:
        raise NotImplementedError

    # # # # # Update # # # # #
    @abstractmethod
    async def update(self, *args, **kwargs) -> None:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def update_one(cls, *args, **kwargs) -> bool:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def update_many(cls, *args, **kwargs) -> int:
        raise NotImplementedError
